/*                                           -*- mode: c; c-basic-offset: 2 -*-
 *
 * This is kiba, a fun GNOME dock
 */

#include <math.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gconf/gconf-client.h>

#include <X11/Xlib.h>

#include "../config.h"
#undef HAVE_GLITZ

#ifdef HAVE_SVG
  #include <librsvg/rsvg.h>
  #include <librsvg/rsvg-cairo.h>
#endif

#ifdef HAVE_GLITZ
  #include <gdk/gdkx.h>
  #include <glitz-glx.h>
  #include <cairo-glitz.h>
#endif

#include "../akamaru/akamaru.h"

#include "kiba-dock.h"
#include "kiba-geometry.h"
#include "kiba-events.h"
#include "kiba-gconf.h"
#include "kiba-object.h"

gboolean
kiba_expose_event (GtkWidget *widget, GdkEventExpose *event)
{
  KibaDock     *dock = KIBA_DOCK (widget);

  if (event->window == widget->window)
    kiba_dock_paint (dock);
  else
  {
    GList      *plugin_list;
    KibaObject *object;

    object = kiba_get_object_for_window (dock, event->window);

    for (plugin_list = dock->plugin_functions; plugin_list != NULL; plugin_list = plugin_list->next)
    {
      PluginFunc *plugin_func = plugin_list->data;

      if (object != NULL && g_strrstr (plugin_func->plugin_name, object->plugin->name))
        kiba_object_paint (dock, object, plugin_func);
    }
  }

  return TRUE;
}

gboolean
kiba_enter_notify_event (GtkWidget *widget, GdkEventCrossing *event)
{
  KibaDock   *dock = KIBA_DOCK (widget);
  KibaObject *object;

  if (event->window != widget->window)
  {
    object = kiba_get_object_for_window (dock, event->window);

    if (object != NULL && !object->dragging && !object->prelight)
    {
      object->prelight      = TRUE;
      dock->prelight_object = object;

//       if (strcmp (dgconf->mouse_over_effect, "zoom") != 0)
        object->rerender = TRUE;
    }
  }

  return TRUE;
}

gboolean
kiba_leave_notify_event (GtkWidget *widget, GdkEventCrossing *event)
{
  KibaDock   *dock = KIBA_DOCK (widget);
  KibaObject *object;

  if (event->window != widget->window)
  {
    object = kiba_get_object_for_window (dock, event->window);

    if (object != NULL && !object->dragging && object->prelight)
    {
      gint mouse_x, mouse_y;

      /* mouse position related to object */
      gdk_window_get_pointer (object->window, &mouse_x, &mouse_y, NULL);

      if (mouse_x > object->size || mouse_x < 0 || mouse_y > object->size || mouse_y < 0)
      {
        object->prelight      = FALSE;
        dock->prelight_object = NULL;

//         if (strcmp (dgconf->mouse_over_effect, "zoom") != 0 ||
//             object->size == dgconf->icon_size)
          object->rerender = TRUE;

        if (dgconf->text_enable && dock->text->showing)
        {
          dock->text->showing = FALSE;
          dock->text->requests++;
          g_timeout_add (dgconf->text_hide_delay, text_timeout, dock);
        }
      }
    }
  }

  return TRUE;
}

gboolean
kiba_scroll_event (GtkWidget *widget, GdkEventScroll  *event)
{
  KibaDock *dock = KIBA_DOCK (widget);
/*
  if ((strcmp (dgconf->object_model, "rope") == 0 ||
       strcmp (dgconf->object_model, "wobbly") == 0) &&
       !dock->dragging_object)
  {

    KibaObject   *object;
    GError       *error = NULL;
    int           new_position, offset;

    object = kiba_get_object_for_window (dock, event->window);
    if (object == NULL)
      return FALSE;

    if (dock->changing_position == TRUE)
      return FALSE;
    dock->changing_position = TRUE;

    if (strcmp (dgconf->object_model, "rope") == 0)
    {
      kiba_delete_rope_strings (dock);
      kiba_delete_rope_spring (dock);
    }
    else
      kiba_delete_wobbly_spring (dock);

    new_position = object->position + (event->direction == GDK_SCROLL_UP ? 1 : -1);
    if (new_position > g_list_length (dock->objects))
      new_position = object->position;
    else if (new_position < 1)
      new_position = 1;

    offset = new_position - object->position;

    while (offset != 0)
    {
      KibaObject *buf_object;

      buf_object = g_list_nth_data (dock->objects, object->position + offset - 1);

      buf_object->position += (offset > 0 ? -1 : 1);
      gconf_client_set_int (dgconf->client,
                            g_strdup_printf ("%s/%s_position", buf_object->key, buf_object->plugin->name),
                            buf_object->position, &error);
      if (error != NULL)
      {
        g_message ("failed to create position key for %s notification: %s",
                    buf_object->name, error->message);
        error = NULL;
      }
      offset += (offset > 0 ? -1 : 1);
    }
    object->position = new_position;
    gconf_client_set_int (dgconf->client,
                          g_strdup_printf ("%s/%s_position", object->key, object->plugin->name),
                          object->position, &error);
    if (error != NULL)
    {
      g_message ("failed to create position key for %s notification: %s",
                  object->name, error->message);
      error = NULL;
    }

    kiba_reorder_objectlist (dock);

    if (strcmp (dgconf->object_model, "rope") == 0)
    {
//       kiba_add_rope_spring (dock, 1.0);
      kiba_add_rope_strings (dock);
    }
    else
      kiba_add_wobbly_spring (dock);

    dock->changing_position = FALSE;
  }
*/
  return TRUE;
}

gboolean
kiba_button_press_event (GtkWidget *widget, GdkEventButton *event)
{
  KibaDock *dock = KIBA_DOCK (widget);
  KibaObject *object = NULL;

  dock->drag_offset_x = event->x;
  dock->drag_offset_y = event->y;
  dock->drag_distance = 0;

  object = kiba_get_object_for_window (dock, event->window);

  if (dock->dragging_object)
    return FALSE;

  if (event->button == 3)
  {
    kiba_create_menu (dock, object, event);

    return TRUE;
  }

  if (widget->window == event->window && event->button == 1)
  {
      dock->dragging_dock = TRUE;
      return TRUE;
  }

  if (object == NULL)
    return FALSE;

  if (event->button == 2 && strcmp (dgconf->object_model, "rope") == 0)
    kiba_delete_rope_strings (dock);
  else if (event->button == 2 && strcmp (dgconf->object_model, "rope") != 0)
    kiba_add_rope_strings (dock);

  /* Remove spring from rope */
  if (strcmp (dgconf->object_model, "rope") == 0)
  {
    kiba_delete_rope_spring (dock);
    if (dgconf->rope_spring_while_dragging == FALSE &&
        event->button != 2)
      kiba_delete_object_spring (dock);
  }

  object->dragging           = TRUE;
  dock->dragging_object      = object;
  dock->mouse_anchor->x      = object->object->position.x - (object->size - object->normal_size) / 2.0;
  dock->mouse_anchor->y      = object->object->position.y - (object->size - object->normal_size) / 2.0;
  dock->mouse_anchor->object = object->object;

  if (dgconf->text_enable && !dgconf->text_while_dragging && dock->text->showing)
  {
    dock->text->showing = FALSE;
    dock->text->requests++;
    g_timeout_add (dgconf->text_hide_delay, text_timeout, dock);
  }

  return TRUE;
}

gboolean
kiba_button_release_event (GtkWidget *widget, GdkEventButton *event)
{
  KibaDock *dock = KIBA_DOCK (widget);

  if (dock->dragging_dock)
  {
    dock->dragging_dock = FALSE;
    return TRUE;
  }

  if (dock->dragging_object && dock->drag_distance <= 10)
  {
    KibaObject  *object;

    /* start exec animation for zooming */
    if (strcmp (dgconf->mouse_over_effect, "zoom") == 0)
      dock->zoom_exec     = TRUE;

    if (dgconf->pulse_exec)
    {
      dock->exec_pulse    = TRUE;
      dock->dragging_object->pulse_exec = TRUE;
      dock->dragging_object->rerender_pulse = TRUE;
    }

    dock->zoom_exec_size  = dock->dragging_object->size;
    dock->exec_object     = dock->dragging_object;

    object = kiba_get_object_for_window (dock, event->window);
    if (object != NULL)
    {
      GList      *plugin_list;

      object->exec = TRUE;

      for (plugin_list = dock->plugin_functions; plugin_list != NULL; plugin_list = plugin_list->next)
      {
        PluginFunc *plugin_func = plugin_list->data;

        if (g_strrstr (plugin_func->plugin_name, object->plugin->name))
          plugin_func->release_func (dock, object);
      }
    }

    /* Bounce the object. */
    if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
      dock->exec_object->object->previous_position.x +=
        (dock->d->halignment - 1) * dgconf->bounce_height;
    else
      dock->exec_object->object->previous_position.y +=
        (dock->d->valignment - 1) * dgconf->bounce_height;
  }
  else if (dock->dragging_object != NULL)
  {
    /* throw the object */
    dock->dragging_object->object->previous_position.x -= dock->dx / dgconf->throw_gravity;
    dock->dragging_object->object->previous_position.y -= dock->dy / dgconf->throw_gravity;

    /* repaint the object without prelight */
    if (!dock->gconf->mouseover_rotating)
    {
      dock->dragging_object->prelight = FALSE;
      dock->dragging_object->rerender = TRUE;
    }
  }

  if (event->button == 2 && (strcmp (dgconf->object_model, "rope") == 0))
    kiba_add_rope_strings (dock);
  if (event->button == 2 && (strcmp (dgconf->object_model, "rope") != 0))
    kiba_delete_rope_strings (dock);

  /* Restore the spring to the rope */
  if (strcmp (dgconf->object_model, "rope") == 0)
  {
    kiba_add_rope_spring (dock, 1.0);
    if (dgconf->rope_spring_while_dragging == FALSE &&
        event->button != 2)
      kiba_add_object_spring (dock);
  }

  if (dock->dragging_object != NULL)
  {
    dock->dragging_object->dragging = FALSE;
    dock->dragging_object           = NULL;
  }

  dock->prelight_object       = NULL;
  dock->mouse_anchor->object  = NULL;

  if (dgconf->text_enable && dgconf->text_while_dragging && dock->text->showing)
  {
    dock->text->showing = FALSE;
    dock->text->requests++;
    g_timeout_add (dgconf->text_hide_delay, text_timeout, dock);
  }

  return TRUE;
}

gboolean
kiba_motion_notify_event (GtkWidget *widget, GdkEventMotion *event)
{
  KibaDock        *dock = KIBA_DOCK (widget);
  GdkModifierType  state;
  int               x, y, new_x, new_y, dx, dy, position;

  gdk_window_get_pointer (gdk_get_default_root_window(), &x, &y, &state);

  if (dock->dragging_dock)
  {
    position = kiba_find_closest_position (dock, x, y);
    gconf_client_set_string  (dgconf->client, KIBA_GCONF_PATH "/options/geometry/position",
                              kiba_layout_data[position].name, NULL);
    return TRUE;
  }

  new_x = x - dock->drag_offset_x + dgconf->icon_size / 2;
  new_y = y - dock->drag_offset_y + dgconf->icon_size / 2;
  dx    = new_x - dock->mouse_anchor->x;
  dy    = new_y - dock->mouse_anchor->y;

  dock->dx = dx;
  dock->dy = dy;

  dock->mouse_anchor->x =
    CLAMP (new_x, dgconf->icon_size / 2, dock->geometry.x +
           dock->geometry.width - dgconf->icon_size /2 );
  dock->mouse_anchor->y =
    CLAMP (new_y, dgconf->icon_size / 2, dock->geometry.y +
           dock->geometry.height - dgconf->icon_size / 2);
  dock->drag_distance += sqrt (dx * dx + dy * dy);

  if (dock->benchmark)
    dock->events++;

  return TRUE;
}

void
kiba_drag_motion (GtkWidget *widget, GdkDragContext *drag_context, gint x, gint y, guint time, gpointer user_data) 
{
  KibaDock *dock = user_data;

  dock->drag_motion = TRUE;

  if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
  {
    dock->drag_geometry.x = x + dock->widget.allocation.x + (0.5 - dock->d->anchor.x) * dock->gconf->icon_size;
    dock->drag_geometry.y = y + dock->widget.allocation.y - dgconf->icon_size;
  }
  else
  {
    dock->drag_geometry.x = x + dock->widget.allocation.x - dgconf->icon_size;
    dock->drag_geometry.y = y + dock->widget.allocation.y + (0.5 - dock->d->anchor.y) * dock->gconf->icon_size;
  }
}

void kiba_drag_leave (GtkWidget *widget, GdkDragContext *drag_context, guint time, gpointer user_data)
{
  KibaDock *dock = user_data;

  dock->drag_motion = FALSE;

  if (dock->drag_window != NULL)
  {
    gdk_window_destroy (dock->drag_window);
    dock->drag_window = NULL;
    dock->drag_window_offset = 0.0;
  }
}

void
kiba_drag_data_received (GtkWidget *widget, GdkDragContext *drag_context, gint x, gint y, GtkSelectionData *selection_data, guint info, guint time)
{
  KibaDock *dock = KIBA_DOCK (widget);
  char     *key;

  gtk_drag_finish (drag_context, TRUE, FALSE, time);
  g_strstrip ((char *) selection_data->data);
  printf ("got desktop file: %s\n", selection_data->data);

  if (selection_data->data[0] != '/' && !strstr ((char *) selection_data->data, "file://"))
  {
    g_message ("Cant handle Relative path\n" "Use a programm for dragging wich sends a fullpath\n"
               "xfce4-appfinder for example should work");
    return;
  }

  key = g_strdup_printf (KIBA_GCONF_PATH "/plugins/launcher/launchers/%s/file",
          g_strndup (g_strrstr ((char *) selection_data->data, "/") + 1,
          strlen (g_strrstr ((char *) selection_data->data, "/") + 1) - 8));

  if (gconf_client_dir_exists (dgconf->client, g_strndup (key, strlen (key) - 5), NULL) == TRUE)
  {
    g_message ("Gconf dir allready exists for launcher %s\n"
               "if u wanna add this launcher twice, please rename the desktop file",
               g_strndup (key, strlen (key) - 5));
    g_free (key);
    return;
  }

  if (strstr ((char *) selection_data->data, "file://"))
    merge_desktop_to_home (dock, (char *) selection_data->data + 7, g_strndup (key, strlen (key) - 5));
  else
    merge_desktop_to_home (dock, (char *) selection_data->data, g_strndup (key, strlen (key) - 5));

  g_free (key);
}
