/*                                           -*- mode: c; c-basic-offset: 2 -*-
 *
 * This is kiba, a fun GNOME dock
 */

#include <string.h>
#include <gtk/gtk.h>
#include <gconf/gconf-client.h>

#include <X11/Xlib.h>

#include "../config.h"
#undef HAVE_GLITZ

#ifdef HAVE_SVG
  #include <librsvg/rsvg.h>
  #include <librsvg/rsvg-cairo.h>
#endif

#ifdef HAVE_GLITZ
  #include <gdk/gdkx.h>
  #include <glitz-glx.h>
  #include <cairo-glitz.h>
#endif

#include "../akamaru/akamaru.h"

#include "kiba-dock.h"
#include "kiba-akamaru.h"
#include "kiba-geometry.h"
#include "kiba-gconf.h"
#include "kiba-object.h"

void
kiba_gconf_init_defaults (AkamaruModel *model, KibaGconf *gconf)
{
  model->elasticity = 0.8;
  model->friction   = 150;
  model->k          = 0.6;

  gconf->memory_object_size    = 80.0;
  gconf->memory_face           = NULL;
  gconf->memory_progress_bg    = NULL;
  gconf->memory_alpha          = 1.0;
  gconf->memory_enable         = TRUE;
  gconf->memory_color          = "#ffffff";
  gconf->memory_border_color   = "#000000";
  gconf->memory_progress_color = "#000000";
  gconf->memory_progress_alpha = 0.4;

  gconf->clock_enable       = TRUE;
  gconf->clock_show_seconds = TRUE;
  gconf->clock_show_date    = TRUE;
  gconf->clock_size = 68.0;
  gconf->clock_alpha = 0.9;
  gconf->clock_face = NULL;
  gconf->clock_border_alpha = 1.0;
  gconf->clock_border_width = 3.5;
  gconf->clock_border_color = "#A2A2A2";
  gconf->clock_color = "#ffffff";
  gconf->clock_date_text_color = "#ffffff";
  gconf->clock_marker_color = "#000000";
  gconf->clock_marker_width = 2.5;
  gconf->clock_circle_radius = 7.0;
  gconf->clock_hour_hand_color = "#000000";
  gconf->clock_minute_hand_color = "#000000";
  gconf->clock_second_hand_color = "#000000";
  gconf->clock_circle_color = "#ffffff";
  gconf->clock_24mode = TRUE;
  gconf->clock_shadow = NULL;
  gconf->clock_marks = NULL;
  gconf->clock_glass = NULL;
  gconf->clock_hour_hand = NULL;
  gconf->clock_minute_hand = NULL;
  gconf->clock_second_hand = NULL;

  gconf->taskbar_icon_alpha     = 0.7;
  gconf->taskbar_icon_size      = 88.0;
  gconf->taskbar_enable         = TRUE;
  gconf->taskbar_border_color   = "#ffffff";
  gconf->taskbar_border_alpha   = 0.8;
  gconf->taskbar_border_radius  = 4.5;
  gconf->taskbar_border_width   = 2.0;
  gconf->taskbar_skip_app       = g_slist_append (gconf->taskbar_skip_app, g_strdup ("cairo-clock"));
  gconf->taskbar_icon_edge         = "bottom_right",
  gconf->taskbar_icon_scale_factor = 0.35;
  gconf->taskbar_inactive_blend_color = "#000000";
  gconf->taskbar_inactive_blend_alpha = 0.5;
  gconf->taskbar_inactive_icon_blend = TRUE;
  gconf->taskbar_fallback_icon = NULL;

  gconf->launcher_running_color      = "#ffffff";
  gconf->launcher_running_alpha      = 0.4;
  gconf->launcher_inactive_alpha     = 0.8;
  gconf->launcher_alpha              = 1.0;
  gconf->icon_size                   = 48;
  gconf->launcher_enable             = TRUE;
  gconf->launcher_multiple_instances =
    g_slist_append (gconf->launcher_multiple_instances, g_strdup ("xterm"));

  gconf->arrow_background = "gnome-dock";
  gconf->animated_background = TRUE;
  gconf->auto_hide = FALSE;
  gconf->auto_hide_area = 100;
  gconf->auto_hide_speed = 25;
  gconf->border_alpha = 1.0;
  gconf->border_color = "#145BC4";
  gconf->border_width = 1.0;
  gconf->bounce_height = 2;
  gconf->cap_size = 0;
  gconf->cap_margin = 0;
  gconf->dock_alpha_1 = 0.9;
  gconf->dock_alpha_2 = 0.0;
  gconf->dock_alpha_3 = 0.9;
  gconf->dock_alpha_4 = 0.9;
  gconf->dock_alpha_5 = 0.9;
  gconf->mouse_translation = 1.0;
  gconf->dock_background = "gnome-dock";
  gconf->dock_bg_affection = 45.0;
  gconf->arrow_bg_affection = 45.0;
  gconf->text_bg_affection = 45.0;
  gconf->dock_bg_step_size = 0.2;
  gconf->text_bg_step_size = 0.2;
  gconf->arrow_bg_step_size = 0.2;
  gconf->dock_bg_radius_1 = 0.2;
  gconf->dock_bg_radius_2 = 0.1;
  gconf->text_bg_radius_1 = 0.2;
  gconf->text_bg_radius_2 = 0.1;
  gconf->arrow_bg_radius_1 = 0.2;
  gconf->arrow_bg_radius_2 = 0.1;
  gconf->dock_color_1 = "#082D76";
  gconf->dock_color_2 = "#1780E1";
  gconf->dock_color_3 = "#082D76";
  gconf->dock_color_4 = "#1780E1";
  gconf->dock_color_5 = "#082D76";
  gconf->dock_corner_bottomleft  = "rounded";
  gconf->dock_corner_bottomright = "rounded";
  gconf->dock_corner_topleft     = "rounded";
  gconf->dock_corner_topright    = "rounded";
  gconf->dock_position_stepsize = 10;
  gconf->exec_rotate = FALSE;
  gconf->exec_rotate_velocity = 5;
  gconf->gravity_multiplier = 60;
  gconf->inertia = 1.5;
  gconf->pulse_exec = TRUE;
  gconf->pulse_exec_zoom_factor = 2.0;
  gconf->pulse_while_dragging = FALSE;
  gconf->pulse_max_alpha = 1.0;
  gconf->object_wobbly_spring = 5000;
  gconf->object_space = 20.0;
  gconf->margin = 55;
  gconf->object_model = "rope";
  gconf->monitor_number = 0;
  gconf->mouse_cursor_related_rotating = FALSE;
  gconf->mouseover_bouncing_height = 0.3;
  gconf->mouseover_rotating = FALSE;
  gconf->mouseover_rotate_velocity = 5;
  gconf->prelight_alpha = 0.4;
  gconf->prelight_color = "#145BC4";
  gconf->pulse_stepsize = 1.5;
  gconf->pulse_exec_stepsize = 3.0;
  gconf->radius = 26;
  gconf->rope_gravity = 500;
  gconf->rope_object_distance = 11.5;
  gconf->rope_spring_while_dragging = FALSE;
  gconf->rotate_area_factor = 2.5;
  gconf->rotate_scale_factor = 0.8;
  gconf->show_all_borders = FALSE;
  gconf->text_alpha = 1.0;
  gconf->text_position_stepsize = 10;
  gconf->text_bg_alpha_1 = 0.9;
  gconf->text_bg_alpha_2 = 0.0;
  gconf->text_bg_alpha_3 = 0.9;
  gconf->text_bg_alpha_4 = 0.9;
  gconf->text_bg_alpha_5 = 0.9;
  gconf->text_background = "gnome-dock";
  gconf->text_bg_color_1 = "#082D76";
  gconf->text_bg_color_2 = "#1780E1";
  gconf->text_bg_color_3 = "#082D76";
  gconf->text_bg_color_4 = "#1780E1";
  gconf->text_bg_color_5 = "#082D76";
  gconf->text_border_alpha = 1.0;
  gconf->text_border_color = "#145BC4";
  gconf->text_border_style = "rounded";
  gconf->text_border_width = 1.0;
  gconf->text_color = "#ffffff";
  gconf->text_hide_delay = 800;
  gconf->text_enable = TRUE;
  gconf->text_icon_padding = 9;
  gconf->text_placement_x = "center";
  gconf->text_placement_y = "center";
  gconf->text_shadow_alpha = 0.8;
  gconf->text_shadow_color = "#ffffff";
  gconf->text_shadow_offset = 2;
  gconf->text_size = 16;
  gconf->text_type = "Sans";
  gconf->text_weight = "normal";
  gconf->text_while_dragging = TRUE;
  gconf->throw_gravity = 2;
  gconf->zoom_max_stepsize = 2.5;
  gconf->zoom_factor = 2.0;
  gconf->zoom_area_factor = 1.0;
  gconf->pulse_zoom_factor = 2.0;
  gconf->mouse_over_effect = "pulse";
}

void
kiba_switch_model (GtkWidget* widget, gpointer data)
{
  GConfClient *gconf_client;
  GError      *error = NULL;
  char        *model = data;

  gconf_client = gconf_client_get_default ();

  gconf_client_set_string  (gconf_client, KIBA_GCONF_PATH "/options/physic/model",
                            model, &error);
  if (error != NULL)
  {
    g_message ("Failed to set option for for " KIBA_GCONF_PATH "/options/physic/model" 
               "notifications: %s", error->message);
    g_error_free (error);
    return;
  }
}

void
toggle_auto_hide (GtkWidget* widget, gpointer data)
{
  KibaDock *dock = data;

  if (dgconf->auto_hide == TRUE)
  {
    if (gconf_client_set_bool (dgconf->client, 
        KIBA_GCONF_PATH "/options/auto_hide/auto_hide", FALSE, NULL) == FALSE)
      dgconf->auto_hide = FALSE;
  }
  else
  {
    if (gconf_client_set_bool (dgconf->client,
        KIBA_GCONF_PATH "/options/auto_hide/auto_hide", TRUE, NULL) == FALSE)
      dgconf->auto_hide = TRUE;
  }
}

void
options_changed_callback (GConfClient *client,
                          guint        cnxn_id,
                          GConfEntry  *entry,
                          gpointer     user_data)
{
  KibaDock *dock;

  if (g_strrstr (entry->key, "/plugins/"))
    return;

  dock = KIBA_DOCK (user_data);

  set_option (dock, entry);

  if (strcmp (entry->key, KIBA_GCONF_PATH "/options/physic/gravity_multiplier") == 0 ||
      strcmp (entry->key, KIBA_GCONF_PATH "/options/geometry/cap_size") == 0 ||
      strcmp (entry->key, KIBA_GCONF_PATH "/options/geometry/radius") == 0 ||
      strcmp (entry->key, KIBA_GCONF_PATH "/options/geometry/margin") == 0)
    kiba_layout (dock);
  else if (g_strrstr (entry->key, "/options/style/dock_alpha_") ||
           g_strrstr (entry->key, "/options/style/dock_color_") ||
           g_strrstr (entry->key, "/options/style/border") ||
           g_strrstr (entry->key, "/options/style/prelight_") ||
           g_strrstr (entry->key, "/options/style/dock_corner_") ||
           strcmp (entry->key, KIBA_GCONF_PATH "/options/geometry/cap_margin") == 0 ||
           strcmp (entry->key, KIBA_GCONF_PATH "/options/style/show_all_borders") == 0 ||
           strcmp (entry->key, KIBA_GCONF_PATH "/options/style/dock_background") == 0 ||
           strcmp (entry->key, KIBA_GCONF_PATH "/options/style/dock_bg_affection") == 0 ||
           strcmp (entry->key, KIBA_GCONF_PATH "/options/style/dock_bg_step_size") == 0 ||
           strcmp (entry->key, KIBA_GCONF_PATH "/options/style/dock_bg_radius_1") == 0 ||
           strcmp (entry->key, KIBA_GCONF_PATH "/options/style/dock_bg_radius_2") == 0)
    dock->rerender = TRUE;
  else if (strcmp (entry->key, KIBA_GCONF_PATH "/options/geometry/position") == 0)
  {
    if (strcmp (dgconf->object_model, "rope") == 0)
    {
      if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
      {
        dock->first_rope_anchor->x = dock->last_rope_anchor->x;
        dock->first_rope_anchor->y = 0.0;
        dock->last_rope_anchor->x  = ((dock->d->halignment - 1 == 1) ? ((double) (dock->geometry.width)) : (0.0));
        dock->last_rope_anchor->y  = (double) dock->geometry.height
                                                    - dgconf->icon_size;
      }
      else
      {
        dock->first_rope_anchor->x = (double) dock->geometry.width
                                              - dgconf->icon_size;
        dock->first_rope_anchor->y = dock->last_rope_anchor->y;
        dock->last_rope_anchor->x  = 0;
        dock->last_rope_anchor->y  = (double) ((dock->geometry.height / 2)
                            + ((dock->d->valignment - 1) * (dock->geometry.height / 2)));
      }

      kiba_delete_rope_spring (dock);
      kiba_add_rope_spring (dock, 1.0);
    }

    dock->rerender = TRUE;

    DO_FOR_ALL_OBJECTS
    (
      if (g_strrstr (object->plugin->name, "memory") ||
          g_strrstr (object->plugin->name, "taskbar"))
        object->rerender = TRUE;
    );
  }
  else if (strcmp (dgconf->object_model, "rope") == 0 && strcmp (entry->key, KIBA_GCONF_PATH "/options/rope_gravity") == 0)
    {
      kiba_delete_rope_spring (dock);
      kiba_add_rope_spring (dock, 1.0);
    }
  else if ((strcmp (dgconf->object_model, "wobbly") == 0 ||
            strcmp (dgconf->object_model, "spring_for_all") == 0) &&
            strcmp (entry->key, KIBA_GCONF_PATH "/options/physic/object_wobbly_spring") == 0)
    {
      kiba_delete_wobbly_spring (dock);
      kiba_add_wobbly_spring (dock);
    }
  else if (!dock->gconf->auto_hide && dock->hiding &&
           strcmp (entry->key, KIBA_GCONF_PATH "/options/auto_hide/auto_hide") == 0)
    {
      dock->hiding = FALSE;
      dock->unhiding = TRUE;
      gtk_widget_show (&dock->widget);
      g_timeout_add (dgconf->auto_hide_speed, kiba_unhide, dock);
    }
  else if (strcmp (entry->key, KIBA_GCONF_PATH "/options/physic/model") == 0 ||
           strcmp (entry->key, KIBA_GCONF_PATH "/options/geometry/monitor_number") == 0)
    kiba_reload (dock);
  else if (strcmp (entry->key, KIBA_GCONF_PATH "/options/physic/rope_object_distance") == 0 &&
           strcmp (dgconf->object_model, "rope") == 0)
    {
      kiba_delete_rope_strings (dock);
      kiba_add_rope_strings (dock);

      kiba_layout (dock);
    }
  else if (strcmp (entry->key, KIBA_GCONF_PATH "/options/text/text_weight") == 0 ||
           strcmp (entry->key, KIBA_GCONF_PATH "/options/text/text_type") == 0 ||
           strcmp (entry->key, KIBA_GCONF_PATH "/options/text/text_size") == 0)
  {
    DO_FOR_ALL_OBJECTS (kiba_text_update_pango_for_object (object, dock->gconf, dock->text););
  }
  else if (strcmp (entry->key, KIBA_GCONF_PATH "/options/physic/object_space") == 0)
  {
    DO_FOR_ALL_OBJECTS
    (
      AddSpacerClosure closure;

      closure.dock           = dock;
      closure.object         = object->object;
      closure.spacer_size    = object->size + dgconf->object_space;
      closure.object->radius = closure.spacer_size / 2.0;

      akamaru_model_for_each_object (&dock->model, add_spacer, &closure);
    );
  }
  else if (strcmp (entry->key, KIBA_GCONF_PATH "/options/mouseover_effects/zoom_factor") == 0)
  {
/*
    DO_FOR_ALL_OBJECTS
    (
      if (object->icon != NULL && load_icon (dock, object, object->icon, TRUE, NULL))
      {
        cairo_t *cr = NULL;

        cr = dock_cairo_create (object);
        if (cr == NULL)
          return;

        fill_surface_buffer (cr, object, TRUE);
        cairo_destroy (cr);
      }
    );
*/
  }
/*
  else if (strcmp (entry->key, KIBA_GCONF_PATH
                   "/options/geometry/spacer_size") == 0)
  {
    GList *g_list;

    for (g_list = dock->spacer_polygons; g_list != NULL; g_list = g_list->next)
    {
      AkamaruPolygon *polygon = g_list->data;
      akamaru_model_delete_polygon (&dock->model, polygon);
    }

    for (g_list = dock->spacer_polygons; g_list != NULL; g_list = g_list->next)
      dock->spacer_polygons = g_list_remove (dock->spacer_polygons, g_list->data);

    kiba_add_spacer_polygons (dock);
  }
*/
}

void
set_option (KibaDock *dock, GConfEntry *entry)
{
  gboolean    model_struct;
  const char *value;
  char       *key;
  int         i;

  define_struct_member (KibaGconf, AkamaruModel);

  if (strcmp (entry->key, KIBA_GCONF_PATH "/options/physic/elasticity") == 0 ||
      strcmp (entry->key, KIBA_GCONF_PATH "/options/physic/friction") == 0 ||
      strcmp (entry->key, KIBA_GCONF_PATH "/options/physic/k") == 0)
    model_struct = TRUE;
  else
    model_struct = FALSE;

  if (strcmp (entry->key, KIBA_GCONF_PATH "/options/geometry/position") == 0 &&
      entry->value != NULL &&
      entry->value->type == GCONF_VALUE_STRING)
    {
      value = gconf_value_get_string (entry->value);

      set_position_from_string (dock, value);
    }
  else for (i = 0; i < G_N_ELEMENTS (options); i++) {
    key = g_strdup_printf ("/%s", options[i].name);

    if (g_str_has_suffix (entry->key, key) &&
	entry->value != NULL &&
	entry->value->type == GCONF_VALUE_FLOAT)
      {
   if (model_struct == FALSE)
	G_STRUCT_MEMBER (double, dgconf,
                    options[i].offset) = gconf_value_get_float (entry->value);
   else
	G_STRUCT_MEMBER (double, &dock->model,
                    options[i].offset) = gconf_value_get_float (entry->value);
      }
    else if (g_str_has_suffix (entry->key, key) &&
	entry->value != NULL &&
	entry->value->type == GCONF_VALUE_STRING)
      {
   if (model_struct == FALSE)
	G_STRUCT_MEMBER (gchar *, dgconf,
                    options[i].offset) = (gchar *) gconf_value_get_string (entry->value);
   else
	G_STRUCT_MEMBER (gchar *, &dock->model,
                    options[i].offset) = (gchar *) gconf_value_get_string (entry->value);
      }
    else if (g_str_has_suffix (entry->key, key) &&
	entry->value != NULL &&
	entry->value->type == GCONF_VALUE_INT)
      {
   if (model_struct == FALSE)
	G_STRUCT_MEMBER (int, dgconf,
                    options[i].offset) = gconf_value_get_int (entry->value);
   else
	G_STRUCT_MEMBER (int, &dock->model,
                    options[i].offset) = gconf_value_get_int (entry->value);
      }
    else if (g_str_has_suffix (entry->key, key) &&
	entry->value != NULL &&
	entry->value->type == GCONF_VALUE_BOOL)
      {
   if (model_struct == FALSE)
	G_STRUCT_MEMBER (gboolean, dgconf,
                    options[i].offset) = gconf_value_get_bool (entry->value);
   else
	G_STRUCT_MEMBER (gboolean, &dock->model,
                    options[i].offset) = gconf_value_get_bool (entry->value);
      }
    else if (g_str_has_suffix (entry->key, key) &&
	entry->value != NULL &&
	entry->value->type == GCONF_VALUE_LIST)
      {
        GSList *buf_list;

        G_STRUCT_MEMBER (GSList*, dgconf, options[i].offset) = NULL;

        for (buf_list = gconf_value_get_list (entry->value); buf_list != NULL; buf_list = buf_list->next)
        {
          GConfValue *list_item = buf_list->data;

          if (list_item != NULL)
            G_STRUCT_MEMBER (GSList*, dgconf, options[i].offset) = g_slist_append (G_STRUCT_MEMBER (GSList*, dgconf, options[i].offset), (char*) gconf_value_get_string (list_item));
        }
      }
    g_free (key);
  }
}

void
kiba_gconf_get_settings (KibaDock *dock)
{
  GError      **error = NULL;
  GSList      *sub_dirs, *dirs, *l;

  /* set all generall options */
  sub_dirs = gconf_client_all_dirs (dgconf->client, KIBA_GCONF_PATH "/options", error);
  if (error != NULL)
    return;

  for (dirs = sub_dirs; dirs != NULL; dirs = dirs->next)
  {
    GSList *options;

    options = gconf_client_all_entries (dgconf->client, dirs->data, error);
    if (error != NULL)
      return;

    for (l = options; l != NULL; l = l->next)
    {
      GConfEntry *entry = l->data;

      set_option (dock, entry);
    }
  }

  /* set all plugin options */
  sub_dirs = gconf_client_all_dirs (dgconf->client, KIBA_GCONF_PATH "/plugins", error);
  if (error != NULL)
    return;

  for (dirs = sub_dirs; dirs != NULL; dirs = dirs->next)
  {
    GSList *options;

    options = gconf_client_all_entries (dgconf->client, dirs->data, error);
    if (error != NULL)
      return;

    for (l = options; l != NULL; l = l->next)
    {
      GConfEntry *entry = l->data;

      set_option (dock, entry);
    }
  }
}
