
#ifndef __KIBA_OBJECT_H__
#define __KIBA_OBJECT_H__

#define PIXMAP_PATH "/usr/share/pixmaps"

#define ZOOM_AREA (dgconf->zoom_area_factor*object->size)
// #define ZOOM_AREA (1.5*object->normal_size*dgconf->zoom_factor)
#define MAX_SIZE (int) object->normal_size*dgconf->zoom_factor
#define PULSE_AREA (1.5 * lgconf->icon_size * lgconf->pulse_zoom_factor)
#define MAX_PULSE_SIZE (dock->exec_pulse == TRUE ? (object->normal_size * dgconf->pulse_exec_zoom_factor) : (object->normal_size*dgconf->pulse_zoom_factor))
#define ROTATE_AREA (dgconf->rotate_area_factor*object->normal_size)
#define ROPE_OBJECT_DISTANCE (dgconf->rope_object_distance / 10)

#define DO_FOR_ALL_OBJECTS(a); \
  GList *g_list; \
  for (g_list = dock->objects; g_list != NULL; g_list = g_list->next) \
  { \
    KibaObject *object = g_list->data; \
    a; \
  }

#define DO_FOR_ALL_PLUGINS(a); \
  GList *g_list; \
  for (g_list = dock->plugins; g_list != NULL; g_list = g_list->next) \
  { \
    KibaPlugin *plugin = g_list->data; \
    a; \
  }

#ifndef __KIBA_DOCK_H__
  typedef struct _KibaObject         KibaObject;
#endif

/* this struct saves a kiba objects, like a launcher, clock... */
struct _KibaObject {
  KibaPlugin               *plugin;
  gint                      plugin_num;

  GdkWindow                *window;
  gboolean                  at_dock;
  gboolean                  dragging;
  gboolean                  exec;
  gboolean                  prelight;   /* true when object is prelighted */
//   gboolean                  position_notify; /* enables disables gconf notify for gconf */
//   int                       position;
  char                     *key;        /* gconf key */
  char                     *name;       /* text for pango rendering */

  /* physic variables */
  AkamaruObject            *object;
  AkamaruSpacer            *spacer;
  AkamaruSpring            *spring;
  AkamaruSpring            *rope_spring;
  AkamaruString            *rope_string;
  GList                    *wobbly_spring;
  double                    velocity_x;
  double                    velocity_y;

  /* rendering related variables */
  GdkPixbuf                *bg_pixbuf;
  GdkPixbuf                *fg_pixbuf;
  GdkPixmap                *pixmap;
  GdkGC                    *gc;
  cairo_surface_t          *background_surface;
  cairo_surface_t          *foreground_surface;
  PangoLayout              *pango_layout;       //saves the pango layout

#ifdef HAVE_GLITZ
  glitz_drawable_format_t  *glitz_draw_format;
  glitz_drawable_t         *glitz_drawable;
  glitz_format_t           *glitz_format;
#endif

#ifdef HAVE_SVG
  RsvgHandle               *bg_svg_handle;
  RsvgHandle               *fg_svg_handle;
#endif

  gboolean                  rerender;
  gboolean                  rerender_pulse;
  gboolean                  rerender_text;
  gboolean                  bgsurface_needs_update;
  gboolean                  fgsurface_needs_update;

  /* pulse effect */
  GdkWindow                *pulse_window;
  GdkPixmap                *pulse_pixmap;
  GdkGC                    *pulse_gc;
  gboolean                  pulse_exec;
  double                    pulse_size;  /* size of the pulse window */
  gint                      pulse_count;

  /* geometry */
  GdkRectangle              win_position;
  GdkRectangle              previous_position;
  double                    drawn_rotate_angle;
  double                    rotate_angle;
  double                    scale_factor; /* scale factor for rotation */
  double                    normal_size; /* gconf size value */
  double                    size;        /* object size */
  double                    buffer_size;
//TODO replace size with height and width
  double                    buffer_width;
  double                    buffer_height;
  double                    width;
  double                    height;
};

void
kiba_object_create_window          (KibaDock *dock, KibaObject *object);
GdkWindow *
kiba_object_create_pulse_window    (KibaObject *object);
KibaObject*
kiba_get_object_for_window         (KibaDock  *dock, GdkWindow *window);
void
kiba_object_timeout                (KibaDock *dock, KibaObject *object, gboolean object_size_modified, gboolean objects_at_dock, gboolean fast_motion);
void
kiba_object_free                   (KibaDock *dock, KibaObject *object);

void
kiba_object_paint                  (KibaDock *dock, KibaObject *object, PluginFunc *plugin_func);
void
kiba_pulse_paint                   (KibaDock *dock, KibaObject *object);
gboolean
kiba_text_paint                    (KibaDock *dock, KibaObject *object, double fade_factor);
void
fill_surface_buffer                (cairo_t* pSourceContext, KibaObject *object, gboolean background);
void
kiba_update_surface_buffer         (KibaObject *object, gboolean background);
gboolean
kiba_load_png                      (KibaDock *dock, KibaObject *object, gchar *icon, GdkPixbuf **pixbuf, GError **error);
#ifdef HAVE_SVG
gboolean
kiba_load_svg                      (KibaObject *object, char* path, RsvgHandle **handle, GError **error);
#endif
void
kiba_text_update_pango_for_object  (KibaObject *object, KibaGconf  *gconf, KibaText   *text);
void
kiba_rotate_object                 (KibaDock *dock, KibaObject *object, cairo_t    *pixmap_context);
void
kiba_zoom_calculate_size           (KibaDock *dock, KibaObject *object, int mouse_l_x, int mouse_l_y, gboolean object_size_modified);

void
kiba_hide_objects                  (KibaDock *dock);
void
kiba_show_objects                  (KibaDock *dock);
static void
kiba_change_object_win_position    (KibaDock *dock, KibaObject *object, gint mouse_x, gint mouse_y);
void
kiba_object_update_position        (KibaDock *dock, KibaObject *object, gboolean reload);
void
kiba_reorder_objectlist            (KibaDock *dock);
void
kiba_update_object_positions       (KibaDock *dock);
void
kiba_object_position_changed       (KibaDock *dock, KibaObject *object, GConfEntry *entry);
void
kiba_update_rope_positions         (KibaDock *dock, KibaObject *object);

void
kiba_object_show_window            (GtkWidget* widget, gpointer data);
void
kiba_object_focus_window           (GtkWidget* widget, gpointer data);
void
kiba_object_iconify_window         (GtkWidget* widget, gpointer data);
void
kiba_object_kill_window            (GtkWidget* widget, gpointer data);
int
kiba_object_get_winstate           (Display *display, GdkWindow *window);
char*
kiba_object_get_wmclass            (Display *display, Window window);
Window
kiba_object_get_active_window      (Display *display);
gboolean
kiba_object_window_urgency         (Display *display, GdkWindow *window);
int
kiba_get_active_workspace_number   (Display *display);
int
kiba_get_win_workspace_number      (Display *display, Window window);

#endif
