/*                                           -*- mode: c; c-basic-offset: 2 -*-
 *
 * This is kiba, a fun GNOME dock
 */

#include <math.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gconf/gconf-client.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <pango/pango.h>
#include <cairo.h>

#include <X11/Xlib.h>

#include "../config.h"
#undef HAVE_GLITZ

#ifdef HAVE_SVG
  #include <librsvg/rsvg.h>
  #include <librsvg/rsvg-cairo.h>
#endif

#ifdef HAVE_GLITZ
  #include <gdk/gdkx.h>
  #include <glitz-glx.h>
  #include <cairo-glitz.h>
#endif

#if GTK_CHECK_VERSION(2,9,0) == FALSE
  #include <X11/extensions/shape.h>
  #include <gdk/gdkx.h>
#endif

#include "../akamaru/akamaru.h"

#include "kiba-dock.h"
#include "kiba-render.h"
#include "kiba-gconf.h"
#include "kiba-object.h"

void
kiba_dock_paint (KibaDock *dock)
{
  GtkWidget       *widget   = GTK_WIDGET (dock);
  cairo_t         *cr       = NULL;
  cairo_pattern_t *pattern;
  gint             x0, y0, x1, y1;
  double           r, g, b;
  const double     radius = dgconf->radius;

  dock->rerender = FALSE;

  /* border points */
  x0 = dgconf->border_width / 2.0;
  x1 = widget->allocation.width  - dgconf->border_width / 2.0;
  y0 = widget->allocation.height - dgconf->border_width / 2.0;
  y1 = dgconf->border_width / 2.0;

  /* update pixmap size */
  if (dock->pixmap != NULL)
  {
    gdk_pixmap_unref (dock->pixmap);
    dock->pixmap = NULL;
  }
  dock->pixmap = create_pixmap (widget->allocation.width,
                                widget->allocation.height);
  if (dock->pixmap == NULL)
    return;

  if (dock->gc)
  {
    gdk_gc_unref (dock->gc);
    dock->gc = NULL;
  }
  dock->gc     = gdk_gc_new (dock->pixmap);
  if (dock->gc == NULL)
    return;

  /* create context on offscreen pixmap */
  cr = gdk_cairo_create (dock->pixmap);
  if (cr == NULL)
    return;
  cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
  cairo_paint (cr);

  cairo_set_operator (cr, CAIRO_OPERATOR_OVER);

  /* draw dock shape */
  kiba_paint_shape (dock, cr,
       dgconf->dock_corner_topleft, dgconf->dock_corner_topright,
       dgconf->dock_corner_bottomleft, dgconf->dock_corner_bottomright,
       TRUE, dgconf->show_all_borders,
       x0, y0, x1, y1, radius, dgconf->border_width);

  /* get background pattern */
  if (g_str_has_suffix (dgconf->dock_background, ".png"))
    pattern = kiba_get_png_pattern (dock, cr, dgconf->dock_background, TRUE);
  else if (strcmp (dgconf->dock_background, "gradient") == 0)
    pattern = kiba_get_gradient_pattern (dock,
                dgconf->dock_color_1, dgconf->dock_color_2,
                dgconf->dock_color_3, dgconf->dock_alpha_1,
                dgconf->dock_alpha_2, dgconf->dock_alpha_3,
                x0, x1, y0, y1);
  else
    pattern = kiba_get_bg_pattern (dock,
                strcmp (dgconf->dock_background, "linear") == 0 ?
                TRUE : FALSE,
                dgconf->dock_color_1, dgconf->dock_color_2,
                dgconf->dock_color_3, dgconf->dock_color_4,
                dgconf->dock_color_5,
                dgconf->dock_alpha_1, dgconf->dock_alpha_2,
                dgconf->dock_alpha_3, dgconf->dock_alpha_4,
                dgconf->dock_alpha_5,
                dgconf->dock_bg_affection,
                dgconf->dock_bg_step_size / 10.0,
                dgconf->dock_bg_radius_1,
                dgconf->dock_bg_radius_2);

  cairo_set_source (cr, pattern);
  cairo_fill_preserve (cr);

  convert_color (dgconf->border_color, &r, &g, &b);
  cairo_set_source_rgba (cr, r, g, b, dgconf->border_alpha);
  cairo_set_line_width (cr, dgconf->border_width);
  cairo_set_line_join (cr, CAIRO_LINE_JOIN_ROUND);
  cairo_stroke (cr);
  cairo_pattern_destroy (pattern);

/*
  cairo_set_source_rgba (cr, 1.0, 0.2, 0.2, 1.0);
  if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
    cairo_arc (cr, x1 - dock->allocation.width/2.0, y0 - dock->allocation.width/2.0,
               0.7*dock->allocation.width/2.0, 0, 2 * M_PI);
  else
    cairo_arc (cr, x1 - dock->allocation.height/2.0, y0 - dock->allocation.height/2.0*0.7,
               0.7*dock->allocation.height/2.0, 0, 2 * M_PI);
  cairo_fill_preserve (cr);
  cairo_stroke (cr);
*/

  /* update window with offscreen pixmap */
  gdk_draw_drawable  (widget->window, dock->gc, dock->pixmap,
                      0, 0, 0, 0, widget->allocation.width, widget->allocation.height);

  /* shape the dock */
/*
  if (dock->last_allocation.width  != widget->allocation.width ||
      dock->last_allocation.height != widget->allocation.height)
    update_input_shape_from_cairo (widget->window, cairo_get_target (cr),
                                   widget->allocation.width, widget->allocation.height, FALSE);
*/
  cairo_destroy (cr);

  dock->old_bg_offset = dock->bg_offset;
  dock->last_allocation.width  = dock->allocation.width;
  dock->last_allocation.height = dock->allocation.height;
}

void
kiba_drag_animation_paint (KibaDock *dock)
{
  cairo_t*  pixmap_context = NULL;
  gint      drag_width, drag_height, fnord, i;
  double    x0, x1, x2, x3, y0, y1, y2, offset, radius;

  if (dock->drag_window == NULL)
    dock->drag_window = kiba_create_drag_window (dock);

  gdk_window_get_size (dock->drag_window, &drag_width, &drag_height);
  gdk_window_move (dock->drag_window, dock->drag_geometry.x, dock->drag_geometry.y);

  /* update pixmap size */
  if (dock->drag_pixmap != NULL)
  {
    gdk_pixmap_unref (dock->drag_pixmap);
    dock->drag_pixmap = NULL;
  }
  dock->drag_pixmap = create_pixmap (drag_width, drag_height);

  if (dock->drag_pixmap == NULL)
    return;

  if (dock->drag_gc)
  {
    gdk_gc_unref (dock->drag_gc);
    dock->drag_gc = NULL;
  }
  dock->drag_gc = gdk_gc_new (dock->drag_pixmap);
  if (dock->gc == NULL)
    return;

  pixmap_context = gdk_cairo_create (dock->drag_pixmap);
  if (pixmap_context == NULL)
    return;

  /* clear context */
  cairo_set_operator (pixmap_context, CAIRO_OPERATOR_CLEAR);
  cairo_paint (pixmap_context);

  /* do the appropriate transformations */
  if (dock->d->anchor.y == 0)
  {
    cairo_translate (pixmap_context, drag_width, drag_height);
    cairo_rotate (pixmap_context, M_PI);
  }
  else if (dock->d->anchor.x == 0)
  {
    cairo_translate (pixmap_context, drag_width, 0);
    cairo_rotate (pixmap_context, M_PI/2);
    /*flip the fields*/
    fnord       = drag_height;
    drag_height = drag_width;
    drag_width  = fnord;
  }
  else if (dock->d->anchor.x == 4)
  {
    cairo_translate (pixmap_context, 0, drag_height);
    cairo_rotate (pixmap_context, 3*M_PI/2);
    /*flip the fields*/
    fnord       = drag_height;
    drag_height = drag_width;
    drag_width  = fnord;
  }

  /* now we can paint along the user space coordinates, thinking we */
  /* would paint downwards */
  cairo_set_operator (pixmap_context, CAIRO_OPERATOR_OVER);
  x0     = 0.325 * drag_width;
  x1     = 0.45 * drag_width;
  x2     = 0.55 * drag_width;
  x3     = 0.675 * drag_width;
  radius = (x1 - x0) / 2.0;
  y0     = radius;
  y1     = 0.2 * drag_height;
  y2     = x2 - x1;

  /* define offset of first arrow */
  offset = dock->drag_window_offset * drag_height;
  for (i = 1; i < 4; i++)
  {
    cairo_pattern_t *pattern;
    double           r, g, b;

    cairo_save (pixmap_context);

    cairo_move_to (pixmap_context, x1, offset + y0);
    cairo_arc_negative (pixmap_context, x0 + radius, offset + y0, radius, 0, M_PI);
    cairo_rel_line_to (pixmap_context,  x1 - x0, y1 - y0);
    cairo_curve_to (pixmap_context, x1 + (x2 - x1) / 2.0, y1 + offset + radius,
                    x1 + (x2 - x1) / 2.0, y1 + offset + radius,
                    x1 + (x2 - x1), y1 + offset);
    cairo_rel_line_to (pixmap_context,  x1 - x0, y0 - y1);
    cairo_arc_negative (pixmap_context, x3 - radius, y0 + offset, radius, 0, M_PI);
    cairo_line_to (pixmap_context, x1 + (x2 - x1) / 2.0, y0 + y2 + offset);
    cairo_line_to (pixmap_context, x1, y0 + offset);
    cairo_close_path (pixmap_context);
    cairo_set_line_join (pixmap_context, CAIRO_LINE_JOIN_ROUND);

    if (g_str_has_suffix (dgconf->arrow_background, ".png"))
      pattern = kiba_get_png_pattern (dock, pixmap_context, dgconf->arrow_background, TRUE);
    else if (strcmp (dgconf->dock_background, "gradient") == 0)
      pattern = kiba_get_gradient_pattern (dock, dgconf->dock_color_1, dgconf->dock_color_2,
                  dgconf->dock_color_3, offset / drag_height, offset / drag_height,
                  offset / drag_height, x0, x1, y0, y1);
    else
      pattern = kiba_get_bg_pattern (dock, strcmp (dgconf->arrow_background, "linear") == 0 ?
                  TRUE : FALSE, dgconf->dock_color_1, dgconf->dock_color_2,
                  dgconf->dock_color_3, dgconf->dock_color_4, dgconf->dock_color_5,
                  offset / drag_height, offset / drag_height, offset / drag_height, offset / drag_height,
                  offset / drag_height, dgconf->arrow_bg_affection, dgconf->arrow_bg_step_size / 10.0,
                  dgconf->arrow_bg_radius_1, dgconf->arrow_bg_radius_2);

    cairo_set_source (pixmap_context, pattern);
    cairo_fill_preserve (pixmap_context);
    cairo_pattern_destroy (pattern);

    convert_color (dgconf->border_color, &r, &g, &b);
    cairo_set_source_rgba (pixmap_context, r, g, b, offset / drag_height);
    cairo_set_line_width (pixmap_context, dgconf->border_width / 2);
    cairo_set_line_join (pixmap_context, CAIRO_LINE_JOIN_ROUND);

    cairo_stroke (pixmap_context);
    cairo_restore (pixmap_context);

    offset = (dock->drag_window_offset + i*0.25) * drag_height;
    if (offset/drag_height > 0.78)
      offset = offset - (0.78 - 0.05) * drag_height;
  }
  dock->drag_window_offset += 0.01;
  if (dock->drag_window_offset >= 0.78)
    dock->drag_window_offset = 0.05;

  /* update window with offscreen pixmap */
  gdk_draw_drawable  (dock->drag_window, dock->drag_gc, dock->drag_pixmap, 0, 0, 0, 0,
          dock->d->orientation == GTK_ORIENTATION_VERTICAL ? drag_height : drag_width,
          dock->d->orientation == GTK_ORIENTATION_VERTICAL ? drag_width : drag_height);

  cairo_destroy (pixmap_context);
}

cairo_pattern_t*
kiba_get_gradient_pattern (KibaDock *dock,
                           gchar *dock_color_1,
                           gchar *dock_color_2,
                           gchar *dock_color_3,
                           double dock_alpha_1,
                           double dock_alpha_2,
                           double dock_alpha_3,
                           double x0, double x1,
                           double y0, double y1)
{
  cairo_pattern_t *pattern;
  gdouble r, g, b, offset;

  if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
    pattern = cairo_pattern_create_linear ((x0 + x1) / 2 - 1, y0,
                                           (x0 + x1) / 2 - 1, y1);
  else
    pattern = cairo_pattern_create_linear (x0, (y0 + y1) / 2,
                                           x1, (y0 + y1) / 2);

  offset = dock->bg_offset / (dock->d->orientation == GTK_ORIENTATION_VERTICAL ?
                                 (double) dock->widget.allocation.height :
                                 (double) dock->widget.allocation.width);

  convert_color (dgconf->dock_color_1, &r, &g, &b);
  cairo_pattern_add_color_stop_rgba (pattern, offset, r, g, b,
                                     dgconf->dock_alpha_1);
  convert_color (dgconf->dock_color_2, &r, &g, &b);
  cairo_pattern_add_color_stop_rgba (pattern, 0.5 + offset, r, g, b,
                                     dgconf->dock_alpha_2);
  convert_color (dgconf->dock_color_3, &r, &g, &b);
  cairo_pattern_add_color_stop_rgba (pattern, 0.9 + offset, r, g, b,
                                     dgconf->dock_alpha_3);

  return pattern;
}

cairo_pattern_t*
kiba_get_png_pattern (KibaDock *dock, cairo_t *cr, gchar *png_file, gboolean rotate)
{
  cairo_pattern_t *pattern;
  cairo_surface_t *image;

  image = cairo_image_surface_create_from_png (png_file);
  if (image == NULL)
  {
    g_message ("failed to load %s\n", png_file);
    png_file = NULL;
    cairo_surface_destroy (image);
    return NULL;
  }

  pattern = cairo_pattern_create_for_surface (image);
  cairo_surface_destroy (image);

  /* rotate cairo context for different dock positions */
  if (rotate)
  {
    if (dock->d->anchor.y == 0)
      cairo_rotate (cr, M_PI);
    else if (dock->d->anchor.x == 0)
      cairo_rotate (cr, M_PI/2);
    else if (dock->d->anchor.x == 4)
      cairo_rotate (cr, 3*M_PI/2);
  }

  cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);

  return pattern;
}

void
kiba_paint_shape (KibaDock *dock,
                  cairo_t  *draw_context,
                  gchar    *topleft_corner,
                  gchar    *topright_corner,
                  gchar    *bottomleft_corner,
                  gchar    *bottomright_corner,
                  gboolean  dock_shape,
                  gboolean  draw_all_border,
                  double    x0,
                  double    y0,
                  double    x1,
                  double    y1,
                  double    radius,
                  double    border_width)
{
  /* move to Top-Left */
  cairo_move_to (draw_context, x0, y1 + radius);

  /* Top-Left Corner */
  if (draw_all_border || dock->d->anchor.y == 4 || dock->d->anchor.x == 4)
  {
    if (strcmp (topleft_corner, "notched") == 0)
      cairo_arc_negative (draw_context, x0, y1,
                          radius, M_PI / 2, 2 * M_PI);
    else if (strcmp (topleft_corner, "mitered") == 0)
      cairo_line_to (draw_context, x0 + radius, y1);
    else if (strcmp (topleft_corner, "rounded") == 0)
      cairo_arc (draw_context, x0 + radius, y1 + radius,
                 radius, M_PI, 3 * M_PI / 2);
    /* squared style */
    else
    {
      cairo_line_to (draw_context, x0, y1);
      cairo_line_to (draw_context, x0 + radius, y1);
    }
  }
  else if (dock->d->anchor.y == 0)
  {
    if (y1 + radius > y0 - radius)
      cairo_move_to (draw_context, x0, y0 - radius);

    cairo_line_to (draw_context, x0, y1 - border_width);
    cairo_line_to (draw_context, x0 + 2*radius, y1 - border_width);
  }
  else if (dock->d->anchor.x == 0)
  {
    cairo_move_to (draw_context, x0 - border_width, y1);
    if (x0 + radius < x1 - radius)
      cairo_line_to (draw_context, x0 + radius, y1);
    else
      cairo_line_to (draw_context, x1 - radius, y1);
  }

  /* Top (left to right) */
  if (dock->d->anchor.y == 4 || (dock->d->anchor.y == 0 && draw_all_border))
    cairo_rel_line_to (draw_context, .0, dgconf->cap_margin);

  if ((dock->d->anchor.x != 4 && dock->d->anchor.x != 0) ||
      (x0 + radius < x1 - radius))
    cairo_rel_line_to (draw_context, x1 - 2*radius - border_width/2, 0);

  if (dock->d->anchor.y == 4 || (dock->d->anchor.y == 0 && draw_all_border))
    cairo_rel_line_to (draw_context, .0, -dgconf->cap_margin);

  /* Top-Right Corner */
  if (draw_all_border || dock->d->anchor.y == 4 ||
      dock->d->anchor.x == 0)
  {
    if (strcmp (topright_corner, "notched") == 0)
      cairo_arc_negative (draw_context, x1, y1,
                          radius, M_PI, M_PI/2);
    else if (strcmp (topright_corner, "mitered") == 0)
    {
      cairo_line_to (draw_context, x1 - radius, y1);
      cairo_line_to (draw_context, x1, y1 + radius);
    }
    else if (strcmp (topright_corner, "rounded") == 0)
      cairo_arc (draw_context, x1 - radius, y1 + radius,
                 radius, 3 * M_PI / 2, 2 * M_PI);
    /* squared style */
    else
    {
      cairo_line_to (draw_context, x1, y1);
      cairo_line_to (draw_context, x1, y1 + radius);
    }
  }
  else if (dock->d->anchor.x == 4)
  {
    cairo_line_to (draw_context, x1 + border_width, y1);
    cairo_line_to (draw_context, x1 + border_width, y1 + 2*radius);
  }
  else if (dock->d->anchor.y == 0 && (y1 + radius < y0 - radius))
    cairo_line_to (draw_context, x1, y1 + radius);

  /* Right (top to bottom) */
  if (dock->d->anchor.x == 0 || (dock->d->anchor.x == 4 && draw_all_border))
    cairo_rel_line_to (draw_context, -dgconf->cap_margin, 0);

  if ((dock->d->anchor.y != 4 && dock->d->anchor.y != 0) ||
      (y1 + radius < y0 - radius) || draw_all_border)
    cairo_rel_line_to (draw_context, 0, y0 - 2*radius - border_width/2);

  if (dock->d->anchor.x == 0 || (dock->d->anchor.x == 4 && draw_all_border))
    cairo_rel_line_to (draw_context, dgconf->cap_margin, 0);

  /* Bottom-Right Corner */
  if (draw_all_border || dock->d->anchor.y == 0 ||
      dock->d->anchor.x == 0)
  {
    if (strcmp (bottomright_corner, "notched") == 0)
      cairo_arc_negative (draw_context, x1, y0,
                          radius, 3 * M_PI / 2, M_PI);
    else if (strcmp (bottomright_corner, "mitered") == 0)
    {
      cairo_line_to (draw_context, x1, y0 - radius);
      cairo_line_to (draw_context, x1 - radius, y0);
    }
    else if (strcmp (bottomright_corner, "rounded") == 0)
      cairo_arc (draw_context, x1 - radius, y0 - radius,
                 radius, 0, M_PI / 2);
    /* squared style */
    else
    {
      cairo_line_to (draw_context, x1, y0);
      cairo_line_to (draw_context, x1 - radius, y0);
    }
  }
  else if (dock->d->anchor.y == 4)
  {
    cairo_line_to (draw_context, x1, y0 + border_width);
    cairo_line_to (draw_context, x1 - 2*radius, y0 + border_width);
  }
  else if (dock->d->anchor.x == 4 && (x0 + radius < x1 - radius))
    cairo_line_to (draw_context, x1 - radius, y0);

  /* Bottom (right to left) */
  if (dock->d->anchor.y == 0 || (dock->d->anchor.y == 4 && draw_all_border))
    cairo_rel_line_to (draw_context, .0, -dgconf->cap_margin);

  if ((dock->d->anchor.x != 4 && dock->d->anchor.x != 0) ||
      (x0 + radius < x1 - radius))
    cairo_rel_line_to (draw_context, -x1 + 2*radius + border_width/2, 0);

  if (dock->d->anchor.y == 0 || (dock->d->anchor.y == 4 && draw_all_border))
    cairo_rel_line_to (draw_context, .0, dgconf->cap_margin);

  /* Bottom-Left Corner */
  if (draw_all_border || dock->d->anchor.y == 0 ||
      dock->d->anchor.x == 4)
  {
    if (strcmp (bottomleft_corner, "notched") == 0)
      cairo_arc_negative (draw_context, x0, y0,
                          radius, 2 * M_PI, 3 * M_PI / 2);
    else if (strcmp (bottomleft_corner, "mitered") == 0)
    {
      cairo_line_to (draw_context, x0 + radius, y0);
      cairo_line_to (draw_context, x0, y0 - radius);
    }
    else if (strcmp (bottomleft_corner, "rounded") == 0)
      cairo_arc (draw_context, x0 + radius, y0 - radius,
                 radius, M_PI / 2, M_PI);
    /* squared style */
    else
    {
      cairo_line_to (draw_context, x0, y0);
      cairo_line_to (draw_context, x0, y0 - radius);
    }
  }
  else if (dock->d->anchor.x == 0)
  {
    cairo_line_to (draw_context, x0 - border_width, y0);
    cairo_line_to (draw_context, x0 - border_width, y0 - 2*radius);
  }
  else if (dock->d->anchor.y == 4 && (y1 + radius < y0 - radius))
    cairo_line_to (draw_context, x0, y0 - radius);

  /* Left (bottom to top) */
  if (dock->d->anchor.x == 4 || (dock->d->anchor.x == 0 && draw_all_border))
    cairo_rel_line_to (draw_context, dgconf->cap_margin, 0);

  if ((dock->d->anchor.y != 4 && dock->d->anchor.y != 0) ||
      (y1 + radius < y0 - radius))
    cairo_rel_line_to (draw_context, 0, -y0 + 2*radius + border_width/2);

  if (dock->d->anchor.x == 4 || (dock->d->anchor.x == 0 && draw_all_border))
    cairo_rel_line_to (draw_context, -dgconf->cap_margin, 0);

  cairo_set_line_join (draw_context, CAIRO_LINE_JOIN_ROUND);
  cairo_close_path (draw_context);
}

cairo_pattern_t*
kiba_get_bg_pattern (KibaDock  *dock,
                     gboolean   linear,
                     gchar     *color1,
                     gchar     *color2,
                     gchar     *color3,
                     gchar     *color4,
                     gchar     *color5,
                     double     alpha1,
                     double     alpha2,
                     double     alpha3,
                     double     alpha4,
                     double     alpha5,
                     double     affection,
                     double     step_size,
                     double     radius_1,
                     double     radius_2)
{
  cairo_pattern_t *pattern;
  cairo_matrix_t   matrix;
  double           step,
                   smaller_side, larger_side,
                   r, g, b;

  if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
  {
    smaller_side  = dock->widget.allocation.width;
    larger_side   = dock->widget.allocation.height;
  }
  else
  {
    smaller_side  = dock->widget.allocation.height;
    larger_side   = dock->widget.allocation.width;
  }

  if (linear)
    pattern = cairo_pattern_create_linear (0.0f, 0.0f,
                  smaller_side * sin ((affection / 180.0) * M_PI),
                  smaller_side * sin ((affection / 180.0) * M_PI - M_PI/2));
  else if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
    pattern =
      cairo_pattern_create_radial
       (smaller_side * 0.5, larger_side * 0.5,
        smaller_side * 0.8 * radius_1,
        smaller_side * 0.5, larger_side * 0.5,
        smaller_side * 0.8 * radius_1);
  else
    pattern =
      cairo_pattern_create_radial
       (larger_side  * 0.5, smaller_side * 0.5,
        smaller_side * 0.8 * radius_1,
        larger_side  * 0.5, smaller_side * 0.5,
        smaller_side * 0.8 * radius_2);

  cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);

  if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
    cairo_matrix_init_translate (&matrix, 10.0f, dock->bg_offset);
  else
    cairo_matrix_init_translate (&matrix, dock->bg_offset, 10.0f);

  cairo_pattern_set_matrix (pattern, &matrix);

  for (step = 0.0f; step < 1.0f; step += step_size)
  {
    convert_color (color1, &r, &g, &b);
    cairo_pattern_add_color_stop_rgba (pattern, step,
                                       r, g, b, alpha1);
    convert_color (color2, &r, &g, &b);
    cairo_pattern_add_color_stop_rgba (pattern,
                                  step + step_size / 4.0,
                                  r, g, b, alpha2);
    convert_color (color3, &r, &g, &b);
    cairo_pattern_add_color_stop_rgba (pattern,
                                  step + step_size / 2.0,
                                  r, g, b, alpha3);
    convert_color (color4, &r, &g, &b);
    cairo_pattern_add_color_stop_rgba (pattern,
                              step + 3.0*step_size / 4.0,
                              r, g, b, alpha4);
  }
  convert_color (color5, &r, &g, &b);
  cairo_pattern_add_color_stop_rgba (pattern, 1.0f, r, g, b, alpha5);

  return pattern;
}

/* Thanks for the shape functions goes to MacSlow, i found them on a
   Example called "input shape" */
#if GTK_CHECK_VERSION(2,9,0) == FALSE
void do_shape_combine_mask (GdkWindow* window,
                            GdkBitmap* mask,
                            gint x,
                            gint y)
{
  Pixmap pixmap;
  int    ignore;
  int    maj;
  int    min;

  if (!XShapeQueryExtension (GDK_WINDOW_XDISPLAY (window), &ignore, &ignore))
    return;

  if (!XShapeQueryVersion (GDK_WINDOW_XDISPLAY (window), &maj, &min))
    return;

//    for shaped input we need at least XShape 1.1
  if (maj != 1 && min < 1)
    return;

  if (mask)
    pixmap = GDK_DRAWABLE_XID (mask);
  else
  {
    x = 0;
    y = 0;
    pixmap = None;
  }

  XShapeCombineMask (GDK_WINDOW_XDISPLAY (window),
                     GDK_DRAWABLE_XID (window),
                     ShapeInput,
                     x,
                     y,
                     pixmap,
                     ShapeSet);
}
#endif

void update_input_shape_from_cairo (GdkWindow *window,
                                    cairo_surface_t *cairo_surface,
                                    int width, int height,
                                    gboolean complete_shape)
{
  GdkBitmap *shape_bitmap = NULL;
  cairo_t   *cr           = NULL;

  shape_bitmap = (GdkBitmap*) gdk_pixmap_new (NULL, width, height, 1);
  if (shape_bitmap)
  {
    cr = gdk_cairo_create (shape_bitmap);
    if (cairo_status (cr) == CAIRO_STATUS_SUCCESS)
    {
      if (complete_shape == FALSE)
      {
        //at least there is no scaling in here, so take care that iWidth and iHeight
        //is exactly the same size as the cairo-context has (at least in pixel)
        cairo_set_source_surface (cr,cairo_surface, 0.0,0.0);
        cairo_set_operator(cr,CAIRO_OPERATOR_SOURCE);
        cairo_paint(cr);
      }
      else
      {
        cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, 0.0);
        cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
        cairo_paint (cr);
      }
      cairo_destroy (cr);
#if GTK_CHECK_VERSION(2,9,0) == FALSE
      do_shape_combine_mask (window, NULL, 0, 0);
      do_shape_combine_mask (window, shape_bitmap, 0, 0);
#else
      gdk_window_input_shape_combine_mask (window, NULL, 0, 0);
      gdk_window_input_shape_combine_mask (window, shape_bitmap, 0, 0);
#endif
    }
  g_object_unref ((gpointer) shape_bitmap);
  }
}

#ifdef HAVE_SVG
void rsvg_resize_func (gint *width, gint *height, gpointer user_data)
{
  KibaObject *object = user_data;

  *width  = (gint) object->buffer_size;
  *height = (gint) object->buffer_size;
}
#endif

GdkPixmap *
create_pixmap (int w, int h)
{
  GdkPixmap   *pixmap;
  GdkVisual   *visual;
  GdkColormap *colormap;

  visual = gdk_visual_get_best_with_depth (32);
  if (!visual)
    return NULL;

  pixmap = gdk_pixmap_new (NULL, w, h, 32);
  if (!pixmap)
    return NULL;

  colormap = gdk_colormap_new (visual, FALSE);
  if (!colormap)
  {
    gdk_pixmap_unref (pixmap);
    return NULL;
  }

  gdk_drawable_set_colormap (GDK_DRAWABLE (pixmap), colormap);
  gdk_colormap_unref (colormap);

  return pixmap;
}

cairo_t *
dock_cairo_create (KibaObject *object)
{
#ifdef HAVE_GLITZ
  if (object->glitz_drawable)
  {
    cairo_t*         cr = NULL;
    cairo_surface_t* cairo_surface;
    glitz_surface_t* glitz_surface;

    glitz_surface = glitz_surface_create (object->glitz_drawable,
         object->glitz_format, object->size,
         object->size, 0, NULL);

    if (glitz_draw_format->doublebuffer)
      glitz_surface_attach (glitz_surface, object->glitz_drawable,
                            GLITZ_DRAWABLE_BUFFER_BACK_COLOR);
    else
      glitz_surface_attach (glitz_surface, object->glitz_drawable,
                            GLITZ_DRAWABLE_BUFFER_FRONT_COLOR);

    cairo_surface = cairo_glitz_surface_create (glitz_surface);
    cr = cairo_create (cairo_surface);

    cairo_surface_destroy (cairo_surface);
    glitz_surface_destroy (glitz_surface);

    return cr;
  }
#endif

  /* update pixmap size */
  if (object->pixmap != NULL)
  {
    gdk_pixmap_unref (object->pixmap);
    object->pixmap = NULL;
  }
  object->pixmap = create_pixmap (object->size, object->size);
  if (object->pixmap == NULL)
    return gdk_cairo_create (object->window);

  if (object->gc)
  {
    gdk_gc_unref (object->gc);
    object->gc = NULL;
  }
  object->gc     = gdk_gc_new (object->pixmap);
  if (object->gc == NULL || object->background_surface == NULL)
    return gdk_cairo_create (object->window);

  return gdk_cairo_create (object->pixmap);
}

/*
void
kiba_polygon_paint (KibaDock *dock)
{
  cairo_t         *cr       = NULL;

  if (dock->polygon_pixmap != NULL)
  {
    gdk_pixmap_unref (dock->polygon_pixmap);
    dock->polygon_pixmap = NULL;
  }
  dock->polygon_pixmap = create_pixmap (dock->geometry.width,
                                dock->geometry.height);
  if (dock->polygon_pixmap == NULL)
    return;

  if (dock->polygon_gc)
  {
    gdk_gc_unref (dock->polygon_gc);
    dock->polygon_gc = NULL;
  }
  dock->polygon_gc     = gdk_gc_new (dock->polygon_pixmap);
  if (dock->polygon_gc == NULL)
    return;

  cr = gdk_cairo_create (dock->polygon_pixmap);
  if (cr == NULL)
    return;

  cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
  cairo_paint (cr);

  cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
  cairo_set_line_join (cr, CAIRO_LINE_JOIN_ROUND);
  cairo_set_line_cap (cr, CAIRO_LINE_CAP_ROUND);

  cairo_set_line_width (cr, 1);
  cairo_set_source_rgba (cr, 0, 0, 0.5, 1);
//   akamaru_model_for_each_polygon (&dock->model, draw_polygon, cr);

  gdk_draw_drawable  (dock->polygon_window,
                      dock->polygon_gc,
                      dock->polygon_pixmap,
                      0, 0, 0, 0,
                      dock->geometry.width,
                      dock->geometry.height);

  update_input_shape_from_cairo (dock->polygon_window,
                                 cairo_get_target (cr),
                                 dock->geometry.width,
                                 dock->geometry.height,
                                 FALSE);
  cairo_destroy (cr);
}
*/

/*
static void
draw_polygon (AkamaruPolygon *polygon, void *data)
{
  cairo_t *cr = data;
  int i;

  for (i = 0; i < polygon->num_points; i++)
    cairo_line_to (cr, polygon->points[i].x, polygon->points[i].y);
  cairo_close_path (cr);

  if (polygon->enclosing)
    cairo_stroke (cr);
  else
    cairo_fill (cr);
}
*/
