/*
  This is gset-kiba a configuration utility for kiba-dock, a fun GNOME dock
*/

#include <gtk/gtk.h>
#include <glib-object.h>
#include <gconf/gconf-client.h>
#include <glade/glade.h>
#include <stdlib.h>
#include <string.h>

/* include the gconf struct from kiba-dock */
#define NO_GCONF_FUNCTS
#include "../dock/kiba-gconf.h"
#undef NO_GCONF_FUNCTS

#define KIBA_GCONF_PATH  "/apps/kiba"
#define PLUGIN_GCONF_PATH  "/apps/kiba/plugins"

#define DO_FOR_GCONF_DIRS(a); \
  GSList *dirs, *plugin_dir, *option_dir; \
  option_dir = gconf_client_all_dirs (client, KIBA_GCONF_PATH "/options", NULL); \
  plugin_dir = gconf_client_all_dirs (client, PLUGIN_GCONF_PATH, NULL); \
  for (dirs = option_dir; dirs != NULL; dirs = dirs->next) \
  { \
    a; \
  } \
  for (dirs = plugin_dir; dirs != NULL; dirs = dirs->next) \
  { \
    a; \
  }

#define GETW(a)          G_OBJECT(glade_xml_get_widget (gset->all_widgets, a))
#define CLICK_EVENT(a)   g_signal_connect (GETW (a), "clicked", G_CALLBACK (on_click_callback), NULL)
#define CVALUE_EVENT(a)  g_signal_connect (GETW (a), "value-changed", G_CALLBACK (on_vchanged_callback), GETW(a))
#define SVALUE_EVENT(a)  g_signal_connect (GETW (a), "value-changed", G_CALLBACK (on_spinbutton_changed_callback), GETW(a))
#define LVALUE_EVENT(a)  g_signal_connect (GETW (a), "activate", G_CALLBACK (on_list_changed_callback), GETW(a))
#define CHANGE_EVENT(a)  g_signal_connect (GETW(a), "changed", G_CALLBACK (on_change_callback), gset)
#define PNG_CHANGE_EVENT(a)  g_signal_connect (GETW(a), "selection-changed", G_CALLBACK (on_png_change_callback), gset)
#define COLOR_EVENT(a)   g_signal_connect (GETW(a), "color-set", G_CALLBACK (on_color_callback), gset)
#define FONT_EVENT(a)    g_signal_connect (GETW(a), "font-set", G_CALLBACK (on_font_change_callback), gset)
#define MODEL(a,b)       G_STRUCT_MEMBER (a, &gset->gconf, b)

typedef struct _GsetKiba   GsetKiba;

struct _GsetKiba {
  KibaGconf       gconf;
  GladeXML       *all_widgets;
  GConfClient    *gconf_client;
  gboolean        color_set;
  gboolean        gconf_event;
  char           *gconf_string;

  /* akamaru engine config values */
  double          friction;
  double          k;
  double          elasticity;
};

void combo_box_get_name (GtkWidget *widget,char *search)
{
  GtkTreeModel  *model;
  GtkTreeIter    iter;
  gboolean       i  = TRUE;

  model = gtk_combo_box_get_model (GTK_COMBO_BOX (widget));

  if (gtk_tree_model_get_iter_first  (model, &iter))
  {
    while(i)
    {
      GValue value = { 0, };
      gtk_tree_model_get_value (model, &iter, 0, &value);
      if(!strcmp (g_value_get_string (&value), search))
        gtk_combo_box_set_active_iter (GTK_COMBO_BOX (widget), &iter);
      i = gtk_tree_model_iter_next (model, &iter);
    }
  }
}

static void
get_option (GsetKiba *gset, GConfEntry *entry)
{
  GtkWidget *widget;
  char      *key;
  int        i;

  define_struct_member (KibaGconf, GsetKiba);

  for (i = 0; i < G_N_ELEMENTS (options); i++)
  {
    key = g_strdup_printf ("/%s", options[i].name);

    /* Get Widget for the gconf-key */
    widget = glade_xml_get_widget (gset->all_widgets, options[i].name);

    /* check if there is an widget for this option */
    if (widget != NULL)
    {
      if (g_str_has_suffix (entry->key, key) &&
          entry->value != NULL &&
          entry->value->type == GCONF_VALUE_FLOAT)
      {
        MODEL (double, options[i].offset) = gconf_value_get_float (entry->value);

        if (g_str_has_suffix (entry->key, "/clock_size") ||
            g_str_has_suffix (entry->key, "/taskbar_icon_size") ||
            g_str_has_suffix (entry->key, "/memory_object_size"))
          gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget),
                                     MODEL (double, options[i].offset));
        else
          gtk_range_set_value (GTK_RANGE (widget), MODEL (double, options[i].offset));
      }
      else if (g_str_has_suffix (entry->key, key) &&
               entry->value != NULL &&
               entry->value->type == GCONF_VALUE_STRING)
      {
        MODEL (gchar*, options[i].offset) =
              (gchar *) gconf_value_get_string (entry->value);

        if (strcmp (options[i].name, "text_type") == 0)
        {
          GConfClient   *gconf_client;
          GError        *error = NULL;
          GSList        *all_options,
                        *l;

          gconf_client = gconf_client_get_default ();

          all_options = gconf_client_all_entries (gconf_client,
                    "/apps/kiba/options/text",
                     &error);
          if (error != NULL)
          {
            g_error ("Failed to get options from " KIBA_GCONF_PATH
                     "notifications: %s", error->message);
            g_error_free (error);
            return;
          }
          for (l = all_options; l != NULL; l = l->next)
          {
            GConfEntry *buf_entry = l->data;

            if (strcmp (buf_entry->key, "/apps/kiba/options/text/text_size") == 0 &&
                buf_entry->value != NULL && buf_entry->value->type == GCONF_VALUE_INT)
              gset->gconf.text_size = gconf_value_get_int (buf_entry->value);
          else if (strcmp (buf_entry->key, "/apps/kiba/options/text/text_weight") == 0 &&
                   buf_entry->value != NULL && buf_entry->value->type == GCONF_VALUE_STRING)
            gset->gconf.text_weight = (gchar *) gconf_value_get_string (buf_entry->value);
          }

          gtk_font_button_set_font_name (GTK_FONT_BUTTON (widget), g_strdup_printf ("%s %s %i",
                         MODEL (gchar*, options[i].offset),
                         MODEL (gchar*, G_STRUCT_OFFSET (KibaGconf, text_weight)),
                         MODEL (gint*,  G_STRUCT_OFFSET (KibaGconf, text_size))));
        }
        else if (g_str_has_suffix (options[i].name, "_position"))
          combo_box_get_name (widget, MODEL (gchar*, options[i].offset));
        else if (strcmp (options[i].name, "text_background") == 0 ||
                  strcmp (options[i].name, "arrow_background") == 0 ||
                  strcmp (options[i].name, "clock_face") == 0 ||
                  strcmp (options[i].name, "clock_shadow") == 0 ||
                  strcmp (options[i].name, "clock_marks") == 0 ||
                  strcmp (options[i].name, "clock_glass") == 0 ||
                  strcmp (options[i].name, "clock_hour_hand") == 0 ||
                  strcmp (options[i].name, "clock_minute_hand") == 0 ||
                  strcmp (options[i].name, "clock_second_hand") == 0 ||
                  strcmp (options[i].name, "memory_face") == 0 ||
                  strcmp (options[i].name, "taskbar_fallback_icon") == 0 ||
                  strcmp (options[i].name, "memory_progress_bg") == 0 ||
                  strcmp (options[i].name, "dock_background") == 0)
        {
          GtkWidget *chooser_widget;

          if (strlen (MODEL (gchar*, options[i].offset)) < 3 ||
              strcmp (MODEL (gchar*, options[i].offset), "linear") == 0 ||
              strcmp (MODEL (gchar*, options[i].offset), "radial") == 0 ||
              strcmp (MODEL (gchar*, options[i].offset), "gradient") == 0)
            return;

          chooser_widget = glade_xml_get_widget (gset->all_widgets,
            strcmp (options[i].name, "text_background") == 0 ?
            "text_bg_png" :
            strcmp (options[i].name, "arrow_background") == 0 ?
            "arrow_bg_png" :
            strcmp (options[i].name, "clock_face") == 0 ?
            "clock_face" :
            strcmp (options[i].name, "clock_glass") == 0 ?
            "clock_glass" :
            strcmp (options[i].name, "clock_hour_hand") == 0 ?
            "clock_hour_hand" :
            strcmp (options[i].name, "clock_minute_hand") == 0 ?
            "clock_minute_hand" :
            strcmp (options[i].name, "clock_second_hand") == 0 ?
            "clock_second_hand" :
            strcmp (options[i].name, "clock_shadow") == 0 ?
            "clock_shadow" :
            strcmp (options[i].name, "clock_marks") == 0 ?
            "clock_marks" :
            strcmp (options[i].name, "memory_face") == 0 ?
            "memory_face" :
            strcmp (options[i].name, "taskbar_fallback_icon") == 0 ?
            "taskbar_fallback_icon" :
            strcmp (options[i].name, "memory_progress_bg") == 0 ?
            "memory_progress_bg" :
            "dock_bg_png");

          if (gtk_file_chooser_get_filename
               (GTK_FILE_CHOOSER (chooser_widget)) == NULL)
            gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (chooser_widget),
                                           MODEL (gchar*, options[i].offset));
          gset->gconf_string = MODEL (gchar*, options[i].offset);
        }
        else
        {
          if (MODEL (gchar*, options[i].offset)[0] == '#')
          {
            GdkColor col;
            int      r,g,b;

            sscanf (MODEL (gchar*, options[i].offset),
                    "#%02X%02X%02X", &r, &g, &b);
            col.red   = (double) r * 256;
            col.green = (double) g * 256;
            col.blue  = (double) b * 256;

            if (!gset->color_set)
            gtk_color_button_set_color ((GtkColorButton *)widget, &col);
            gset->color_set = FALSE;
          }
          else
          {
            combo_box_get_name (widget, MODEL (gchar*, options[i].offset));
            gset->gconf_string = MODEL (gchar*, options[i].offset);
            if(!gtk_combo_box_get_active_text (GTK_COMBO_BOX (widget)))
              gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 0);
          }
        }
      }
      else if (g_str_has_suffix (entry->key, key) &&
               entry->value != NULL &&
               entry->value->type == GCONF_VALUE_INT)
      {
        MODEL (int, options[i].offset) = gconf_value_get_int (entry->value);

        if (!g_str_has_suffix (entry->key, "/clock/position") &&
            (g_str_has_suffix (entry->key, "/icon_size") || g_str_has_suffix (entry->key, "position")))
          gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), MODEL (int, options[i].offset));
        else if (!g_str_has_suffix (entry->key, "/clock/position"))
          gtk_range_set_value (GTK_RANGE (widget), MODEL (int, options[i].offset));
      }
      else if (g_str_has_suffix (entry->key, key) &&
               entry->value != NULL &&
               entry->value->type == GCONF_VALUE_BOOL)
      {
        MODEL (gboolean, options[i].offset) = gconf_value_get_bool (entry->value);
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget),
                                      MODEL (gboolean, options[i].offset));
      }
      else if (g_str_has_suffix (entry->key, key) &&
               entry->value != NULL &&
               entry->value->type == GCONF_VALUE_LIST)
      {
        GSList *l, *buf_list = NULL;
        gchar *exclude_str = NULL, *buf = NULL;
        buf_list = gconf_value_get_list (entry->value);
        MODEL (GSList*, options[i].offset) = buf_list;

        for (l = buf_list; l != NULL; l = l->next)
        {
          GConfValue *value = l->data;
          gchar      *exclude_string = NULL;

          exclude_string = g_strdup ((char*) gconf_value_get_string (value));

          if (buf != NULL)
            exclude_str = g_strdup_printf ("%s, %s", buf, exclude_string);
          else
            exclude_str = g_strdup (exclude_string);

          buf = g_strdup (exclude_str);

          g_free (exclude_string);
        }

        gtk_entry_set_text (GTK_ENTRY (widget), exclude_str);
        g_free (exclude_str);
      }
      g_free (key);
    }
  }
}

void
on_click_callback (GtkButton *check_button, gpointer user_data)
{
  GConfClient  *client;
  GSList       *l;
  gboolean      is_set;

  client   = gconf_client_get_default ();

  DO_FOR_GCONF_DIRS
  (
    GSList *options;

    options = gconf_client_all_entries (client, dirs->data, NULL);

    for (l = options; l != NULL; l = l->next)
    {
      GConfEntry *entry = l->data;

      if (g_str_has_suffix (entry->key, g_strdup_printf ("/%s",
                            gtk_widget_get_name (GTK_WIDGET (check_button)))))
      {
        is_set = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (check_button));
        gconf_client_set_bool (client, entry->key, is_set, NULL);
      }
    }
  );
}

void
on_vchanged_callback (GtkRange *range, gpointer user_data)
{
  GtkWidget    *widget;
  GConfClient  *client;
  GSList       *l;

  widget   = user_data;
  client   = gconf_client_get_default ();


  DO_FOR_GCONF_DIRS
  (
    GSList *options;

    options = gconf_client_all_entries (client, dirs->data, NULL);

    for (l = options; l != NULL; l = l->next)
    {
      GConfEntry *entry = l->data;

      if (g_str_has_suffix (entry->key, g_strdup_printf ("/%s",
                            gtk_widget_get_name (widget))))
      {
        float the_value;

        the_value = gtk_range_get_value (range);

        GConfValue *Value = gconf_client_get (client, entry->key, NULL);
        switch (Value->type)
        {
          case GCONF_VALUE_INT:
            gconf_client_set_int (client, entry->key, (int) the_value, NULL);
            break;
          case GCONF_VALUE_FLOAT:
            gconf_client_set_float (client, entry->key, the_value, NULL);
            break;
          default:
            printf ("Error: Wrong type in on_vchanged_callback()\n");
            break;
        }
      }
    }
  );
}

void
on_spinbutton_changed_callback (GtkSpinButton *spinbutton, gpointer user_data)
{
  GtkWidget    *widget;
  GConfClient  *client;
  GSList       *l;

  widget   = user_data;
  client   = gconf_client_get_default ();

  DO_FOR_GCONF_DIRS
  (
    GSList *options;

    options = gconf_client_all_entries (client, dirs->data, NULL);

    for (l = options; l != NULL; l = l->next)
    {
      GConfEntry *entry = l->data;

      if (g_str_has_suffix (entry->key, g_strdup_printf ("/%s",
                            gtk_widget_get_name (widget))))
      {
        switch (entry->value->type)
        {
        case GCONF_VALUE_INT:
        {
            int int_value;
            int_value = gtk_spin_button_get_value_as_int (spinbutton);

            if (g_str_has_suffix (entry->key, "/clock_position"))
              gconf_client_set_int (client, "/apps/kiba/plugins/clock/position", int_value, NULL);
            else
              gconf_client_set_int (client, entry->key, int_value, NULL);
          break;
        }
        case GCONF_VALUE_FLOAT:
        {
            double float_value;
            float_value = gtk_spin_button_get_value_as_float (spinbutton);
            gconf_client_set_float (client, entry->key, float_value, NULL);
          break;
        }
        }
      }
    }
  );
}

void
on_list_changed_callback (GtkComboBox *combobox, gpointer user_data)
{
  GtkWidget    *widget;
  GConfClient  *client;
  GSList       *l;

  widget   = user_data;
  client   = gconf_client_get_default ();

  DO_FOR_GCONF_DIRS
  (
    GSList *options;

    options = gconf_client_all_entries (client, dirs->data, NULL);

    for (l = options; l != NULL; l = l->next)
    {
      GConfEntry *entry = l->data;

      if (g_str_has_suffix (entry->key, g_strdup_printf ("/%s", gtk_widget_get_name (widget))))
      {
        GSList *new_list = NULL;
        const gchar *entry_text = gtk_entry_get_text (GTK_ENTRY (widget));
        gchar *buf;

        while (buf != NULL)
        {
          GConfValue *value;
          value = gconf_value_new (GCONF_VALUE_STRING);

          gchar *str_buf = g_strrstr (entry_text, ",");

          if (str_buf != NULL)
          {
            buf = g_strrstr (entry_text, ",") + 2;
            entry_text = g_strndup (entry_text, strlen(entry_text) - strlen (buf) - 2);
            gconf_value_set_string (value, buf);

            new_list = g_slist_prepend (new_list, value);
          }
          else
          {
            gconf_value_set_string (value, entry_text);
            new_list = g_slist_prepend (new_list, value);

            gconf_value_set_list (entry->value, new_list);
            gconf_client_set (client, entry->key, entry->value, NULL);

            break;
          }
        }
      }
    }
  );
}

void
on_color_callback (GtkColorButton *widget, gpointer user_data)
{
  GsetKiba     *gset;
  GdkColor      col;
  GConfClient  *client;
  GSList       *l;

  gset     = user_data;
  client   = gconf_client_get_default ();

  DO_FOR_GCONF_DIRS
  (
    GSList *options;

    options = gconf_client_all_entries (client, dirs->data, NULL);

    for (l = options; l != NULL; l = l->next)
    {
      GConfEntry *entry = l->data;

      if (g_str_has_suffix (entry->key, g_strdup_printf ("/%s",
                            gtk_widget_get_name (GTK_WIDGET (widget)))))
      {
        char buffer[7];

        gtk_color_button_get_color (widget, &col);

        sprintf (buffer, "#%02X%02X%02X",
                 col.red/256, col.green/256, col.blue/256);

        gconf_client_set_string (client, entry->key, buffer, NULL);
        gset->color_set = TRUE;
      }
    }
  );
}


void
on_change_callback (GtkComboBox *combobox, gpointer user_data)
{
  GsetKiba     *gset;
  GConfClient  *client;
  GSList       *l;

  if (gtk_combo_box_get_active_text (combobox) == NULL)
    return;

  gset     = user_data;
  client   = gconf_client_get_default ();

  DO_FOR_GCONF_DIRS
  (
    GSList *options;

    options = gconf_client_all_entries (client, dirs->data, NULL);

    for (l = options; l != NULL; l = l->next)
    {
      GConfEntry *entry = l->data;

      if (g_str_has_suffix (entry->key, g_strdup_printf ("/%s",
                            gtk_widget_get_name (GTK_WIDGET (combobox)))))
      {
        const gchar *text = (const gchar *) gtk_combo_box_get_active_text (combobox);

        /* clear file_chooser button */
        if (g_str_has_suffix (entry->key, "text_background") ||
            g_str_has_suffix (entry->key, "arrow_background") ||
            g_str_has_suffix (entry->key, "dock_background") ||
            g_str_has_suffix (entry->key, "clock_face") ||
            g_str_has_suffix (entry->key, "clock_glass") ||
            g_str_has_suffix (entry->key, "clock_hour_hand") ||
            g_str_has_suffix (entry->key, "clock_minute_hand") ||
            g_str_has_suffix (entry->key, "clock_second_hand") ||
            g_str_has_suffix (entry->key, "clock_shadow") ||
            g_str_has_suffix (entry->key, "clock_marks") ||
            g_str_has_suffix (entry->key, "memory_face") ||
            g_str_has_suffix (entry->key, "memory_progress_bg"))
        {
          GtkWidget   *file_chooser = NULL;
          file_chooser = glade_xml_get_widget (gset->all_widgets,
            strcmp (gtk_widget_get_name (GTK_WIDGET (combobox)), "text_background") == 0 ?
            "text_bg_png" :
            strcmp (gtk_widget_get_name (GTK_WIDGET (combobox)), "arrow_background") == 0 ?
            "text_bg_png" :
            strcmp (gtk_widget_get_name (GTK_WIDGET (combobox)), "clock_face") == 0 ?
            "clock_face" :
            strcmp (gtk_widget_get_name (GTK_WIDGET (combobox)), "clock_glass") == 0 ?
            "clock_glass" :
            strcmp (gtk_widget_get_name (GTK_WIDGET (combobox)), "clock_hour_hand") == 0 ?
            "clock_hour_hand" :
            strcmp (gtk_widget_get_name (GTK_WIDGET (combobox)), "clock_minute_hand") == 0 ?
            "clock_minute_hand" :
            strcmp (gtk_widget_get_name (GTK_WIDGET (combobox)), "clock_second_hand") == 0 ?
            "clock_second_hand" :
            strcmp (gtk_widget_get_name (GTK_WIDGET (combobox)), "clock_shadow") == 0 ?
            "clock_shadow" :
            strcmp (gtk_widget_get_name (GTK_WIDGET (combobox)), "clock_marks") == 0 ?
            "clock_marks" :
            strcmp (gtk_widget_get_name (GTK_WIDGET (combobox)), "memory_face") == 0 ?
            "memory_face" :
            strcmp (gtk_widget_get_name (GTK_WIDGET (combobox)), "taskbar_fallback_icon") == 0 ?
            "taskbar_fallback_icon" :
            strcmp (gtk_widget_get_name (GTK_WIDGET (combobox)), "memory_progress_bg") == 0 ?
            "memory_progress_bg" :
            "dock_bg_png");
          if (file_chooser != NULL)
            gtk_file_chooser_select_filename (GTK_FILE_CHOOSER (file_chooser), "/");
        }

        gconf_client_set_string (client, entry->key, text, NULL);
      }
    }
  );
}

void
on_png_change_callback (GtkFileChooser *chooser, gpointer user_data)
{
  GsetKiba     *gset;
  GConfClient  *client;
  GSList       *l;
  gchar        *text = NULL;
  GtkWidget    *widget;

  if (gtk_file_chooser_get_filename (chooser) == NULL)
    return;

  gset     = user_data;
  client   = gconf_client_get_default ();

  widget = glade_xml_get_widget (gset->all_widgets,
    strcmp (gtk_widget_get_name (GTK_WIDGET (chooser)), "text_bg_png") == 0 ?
    "text_background" :
    strcmp (gtk_widget_get_name (GTK_WIDGET (chooser)), "arrow_bg_png") == 0 ?
    "arrow_background" :
    strcmp (gtk_widget_get_name (GTK_WIDGET (chooser)), "clock_face") == 0 ?
    "clock_face" :
    strcmp (gtk_widget_get_name (GTK_WIDGET (chooser)), "clock_glass") == 0 ?
    "clock_glass" :
    strcmp (gtk_widget_get_name (GTK_WIDGET (chooser)), "clock_hour_hand") == 0 ?
    "clock_hour_hand" :
    strcmp (gtk_widget_get_name (GTK_WIDGET (chooser)), "clock_minute_hand") == 0 ?
    "clock_minute_hand" :
    strcmp (gtk_widget_get_name (GTK_WIDGET (chooser)), "clock_second_hand") == 0 ?
    "clock_second_hand" :
    strcmp (gtk_widget_get_name (GTK_WIDGET (chooser)), "clock_shadow") == 0 ?
    "clock_shadow" :
    strcmp (gtk_widget_get_name (GTK_WIDGET (chooser)), "clock_marks") == 0 ?
    "clock_marks" :
    strcmp (gtk_widget_get_name (GTK_WIDGET (chooser)), "memory_face") == 0 ?
    "memory_face" :
    strcmp (gtk_widget_get_name (GTK_WIDGET (chooser)), "taskbar_fallback_icon") == 0 ?
    "taskbar_fallback_icon" :
    strcmp (gtk_widget_get_name (GTK_WIDGET (chooser)), "memory_progress_bg") == 0 ?
    "memory_progress_bg" :
    "dock_background");

  DO_FOR_GCONF_DIRS
  (
    GSList *options;

    options = gconf_client_all_entries (client, dirs->data, NULL);

    for (l = options; l != NULL; l = l->next)
    {
      GConfEntry *entry = l->data;

      if (g_str_has_suffix (entry->key, g_strdup_printf ("/%s",
                            gtk_widget_get_name (widget))))
      {
        const gchar *text = (const gchar *) gtk_file_chooser_get_filename (chooser);

        gtk_combo_box_set_active (GTK_COMBO_BOX (widget) , -1);

        if (text != NULL)
          gconf_client_set_string (client, entry->key, text, NULL);
      }
    }
  );
}

void
on_font_change_callback (GtkFontButton *font_button, gpointer user_data)
{
  GsetKiba     *gset;
  GConfClient  *client;
  GSList       *sub_dirs,
               *dirs,
               *l;

  gset     = user_data;
  client   = gconf_client_get_default ();

  sub_dirs = gconf_client_all_dirs (client, KIBA_GCONF_PATH "/options", NULL);

  for (dirs = sub_dirs; dirs != NULL; dirs = dirs->next)
  {
    GSList *options;

    options = gconf_client_all_entries (client, dirs->data, NULL);

    for (l = options; l != NULL; l = l->next)
    {
      GConfEntry *entry = l->data;

      if (g_str_has_suffix (entry->key, g_strdup_printf ("/%s",
                            gtk_widget_get_name (GTK_WIDGET (font_button)))))
      {
        gchar *font_data  =
          (gchar *) gtk_font_button_get_font_name (font_button);
        gchar *font_name   = g_strdup ("_"),
              *font_buf    = g_strdup ("_"),
              *font_weight = g_strdup ("_");
        gint   font_size   = 0;

        /* get font name */
        gint count = 1;
        while (g_strrstr (font_name, " ") == NULL)
        {
          font_name = g_strndup (font_data, count);
          count++;
        }

        font_data = g_strdup (font_data) + strlen (font_name);
        count = 1;
        while (g_strrstr (font_buf, " ") == NULL &&
               strlen (font_data) > count - 1)
        {
          font_buf = g_strndup (font_data, count);
          count++;
        }

        /* get fontWeight and fontSize */
        font_size = atoi (font_buf);
        if (font_size == 0)
        {
          font_weight = g_strdup (font_buf);
          font_size = atoi (g_strdup (font_data) + strlen (font_weight));
        }
        else
          font_weight = NULL;
        g_free (font_buf);

        if (strcmp (gset->gconf_string, font_name))
        {
          gconf_client_set_string (client, entry->key, font_name, NULL);
          gset->gconf_string = (gchar *) font_name;
          if (font_size != 0)
            gconf_client_set_int (client, "/apps/kiba/options/text/text_size", font_size, NULL);
          if (font_weight != NULL)
            gconf_client_set_string (client, "/apps/kiba/options/text/text_weight", font_weight, NULL);
        }
        g_free (font_name);
        if (font_weight != NULL)
          g_free (font_weight);
      }
    }
  }
}

static void
options_changed_callback (GConfClient *client,
                          guint cnxn_id,
                          GConfEntry *entry,
                          gpointer user_data)
{
  GsetKiba     *gset;

  gset = user_data;
  get_option (gset, entry);
}

static void
init_from_gconf (GsetKiba *gset, GConfClient *client, GError **error)
{
  GSList *launchers, *l;

  launchers = gconf_client_all_dirs (client,
                 KIBA_GCONF_PATH "/plugins/launcher/launchers", error);
  if (*error != NULL)
    return;

  DO_FOR_GCONF_DIRS
  (
    GSList *options;

    options = gconf_client_all_entries (client, dirs->data, error);
    if (*error != NULL)
      return;

    for (l = options; l != NULL; l = l->next)
    {
      GConfEntry *entry = l->data;

      get_option (gset, entry);
    }
  );
}

static void
update_preview_cb (GtkFileChooser *file_chooser, gpointer data)
{
  GtkWidget   *preview;
  GdkPixbuf   *pixbuf;
  gboolean     have_preview;
  char        *filename;

  preview      = GTK_WIDGET (data);
  filename     = gtk_file_chooser_get_preview_filename (file_chooser);
  pixbuf       = gdk_pixbuf_new_from_file_at_size (filename, 128, 128, NULL);
  g_free (filename);
  have_preview = (pixbuf != NULL);

  gtk_image_set_from_pixbuf (GTK_IMAGE (preview), pixbuf);
  if (pixbuf)
    g_object_unref (pixbuf);

  gtk_file_chooser_set_preview_widget_active (file_chooser, have_preview);
}

static void
setup_callbacks (GsetKiba *gset)
{
  GConfClient *client;
  GConfValue  *value;
  GSList      *l;

  client   = gconf_client_get_default ();

  DO_FOR_GCONF_DIRS
  (
    GSList *options;

    options = gconf_client_all_entries (client, dirs->data, NULL);

    for (l = options; l != NULL; l = l->next)
    {
      GConfEntry  *entry  = l->data;
      GObject     *widget = NULL;
      char        *name;

      name  = g_strdup (g_strrstr (entry->key, "/") + 1);
      value = gconf_client_get (client, entry->key, NULL);

      /* get widget for this entry */
      widget = GETW (name);

      if (widget == NULL)
      {
/*
        if (g_strrstr (entry->key, "text_size")   == NULL &&
            g_strrstr (entry->key, "text_weight") == NULL)
        printf ("missing widget for gconf key %s\n", entry->key);
*/
      }
      else if (value != NULL)
      {
        switch (value->type)
        {
          case GCONF_VALUE_INVALID:
            printf ("Error: Invalid value in set_callbacks()\n");
            break;
          case GCONF_VALUE_STRING:
          {
            const char *string = (gchar *) gconf_value_get_string (value);

            if (g_strrstr (entry->key, "text_type"))
              FONT_EVENT (name);
            else if (g_strrstr (entry->key, "_color"))  //If string is a Color
              COLOR_EVENT (name);
            //If string is a ComboBox
            else if (!g_strrstr (entry->key, "_face") &&
                     !g_strrstr (entry->key, "memory_progress_bg"))
              CHANGE_EVENT (name);
            break;
          }
          case GCONF_VALUE_INT:
          case GCONF_VALUE_FLOAT:
            if (g_str_has_suffix (entry->key, "/icon_size") ||
                g_str_has_suffix (entry->key, "/clock_size") ||
                g_str_has_suffix (entry->key, "/taskbar_icon_size") ||
                g_str_has_suffix (entry->key, "/clock_position") ||
                g_str_has_suffix (entry->key, "/memory_object_size") ||
                g_str_has_suffix (entry->key, "/memory_position"))
              SVALUE_EVENT (name);
            else if (!g_str_has_suffix (entry->key, "/clock/position"))
              CVALUE_EVENT (name);
            break;
          case GCONF_VALUE_BOOL:
            CLICK_EVENT (name);
            break;
          case GCONF_VALUE_LIST:
            LVALUE_EVENT (name);
            break;
          case GCONF_VALUE_SCHEMA:
          case GCONF_VALUE_PAIR:
            break;
        }
      }

      g_free (name);
    }
  );

  PNG_CHANGE_EVENT ("text_bg_png");
  PNG_CHANGE_EVENT ("arrow_bg_png");
  PNG_CHANGE_EVENT ("clock_face");
  PNG_CHANGE_EVENT ("clock_glass");
  PNG_CHANGE_EVENT ("clock_hour_hand");
  PNG_CHANGE_EVENT ("clock_minute_hand");
  PNG_CHANGE_EVENT ("clock_second_hand");
  PNG_CHANGE_EVENT ("clock_shadow");
  PNG_CHANGE_EVENT ("clock_marks");
  PNG_CHANGE_EVENT ("memory_face");
  PNG_CHANGE_EVENT ("memory_progress_bg");
  PNG_CHANGE_EVENT ("taskbar_fallback_icon");
  PNG_CHANGE_EVENT ("dock_bg_png");
}

void
hook_up_callbacks_and_set_defaults (GsetKiba *gset)
{
  GtkWidget    *widget;
  GConfClient  *client;
  GError       *error = NULL;

  client = gconf_client_get_default ();

  /* Get the main_window and connect the close button to gtk_main_quit */
  widget = glade_xml_get_widget (gset->all_widgets, "Kiba Settings Editor");
  g_signal_connect (G_OBJECT (widget), "delete_event", G_CALLBACK (gtk_main_quit), &error);
  if (error != NULL)
    g_error ("Failed to listen for " KIBA_GCONF_PATH
             "/options notifications: %s", error->message);
  g_clear_error (&error);

  init_from_gconf (gset, client, &error);
  if (error != NULL)
    g_error ("Failed to read config from gconf: %s", error->message);
  g_clear_error (&error);

  gconf_client_add_dir (client, KIBA_GCONF_PATH, GCONF_CLIENT_PRELOAD_NONE, &error);

  /* add notify for all subdirs in /apps/kiba */
  DO_FOR_GCONF_DIRS
  (
    gconf_client_notify_add (client, dirs->data, options_changed_callback, gset, NULL, &error);
    if (error != NULL)
    {
      g_error ("Failed to listen for %s notifications: %s",
               (char *) dirs->data, error->message);
      g_error_free (error);
      exit (1);
    }
  );

  setup_callbacks (gset);
}

int main (int argc, char *argv[])
{
  GsetKiba *gset;

  gset = malloc (sizeof (GsetKiba));

  gtk_init (&argc, &argv);
  gconf_init(argc, argv, NULL);

  /* load the interface */
  if (g_file_test ("gset-kiba.glade", (G_FILE_TEST_EXISTS)))
    gset->all_widgets =
      glade_xml_new ("gset-kiba.glade", NULL, NULL);
  else if (g_file_test ("gset-kiba/gset-kiba.glade", (G_FILE_TEST_EXISTS)))
    gset->all_widgets =
      glade_xml_new ("gset-kiba/gset-kiba.glade", NULL, NULL);
  else if (g_file_test ("../gset-kiba/gset-kiba.glade", (G_FILE_TEST_EXISTS)))
    gset->all_widgets =
      glade_xml_new ("../gset-kiba/gset-kiba.glade", NULL, NULL);
  else if (g_file_test ("/usr/share/kiba-dock/gset-kiba.glade",
          (G_FILE_TEST_EXISTS)))
    gset->all_widgets =
      glade_xml_new ("/usr/share/kiba-dock/gset-kiba.glade", NULL, NULL);
  else if (g_file_test ("/usr/local/share/kiba-dock/gset-kiba.glade",
          (G_FILE_TEST_EXISTS)))
    gset->all_widgets =
      glade_xml_new ("/usr/local/share/kiba-dock/gset-kiba.glade", NULL, NULL);
  else
  {
    printf ("cant find glade file, run \"make install\""
            "or change to the kiba source dir\n");
    return 1;
  }

  gset->gconf_event = FALSE;
  /* Connect all the signals to the appropriate callbacks */
  hook_up_callbacks_and_set_defaults (gset);

  /* start the event loop */
  gtk_main ();

  free (gset);
  return 0;
}
